import copy
import json
import os
from collections import deque
from typing import Any
from typing import Dict

import click
import inquirer
import jinja2

from synalinks.cli.constants import PROJECT_CONFIG_FILENAME
from synalinks.cli.constants import PROJECT_STRUCTURE_TEMPLATE
from synalinks.cli.utils.project_utils import get_synalinks_project_config
from synalinks.cli.utils.project_utils import is_inside_synalinks_project
from synalinks.src.utils.naming import to_pkg_name
from synalinks.src.version import version as get_version


def get_config_json(filename: str):
    """
    Reads and returns the JSON configuration from a specified file within the config
    folder.

    This function locates the configuration folder using the `synalinks.cli.config`
    module, opens the specified JSON file, and parses its contents into a Python
    dictionary.

    Args:
        filename (str): The name of the JSON configuration file to read.

    Returns:
        (dict): A dictionary containing the JSON configuration data.
    """
    from synalinks.cli import config

    config_folder_path = config.__file__[: -len("/__init__.py")]
    json_config = {}
    with open(os.path.join(config_folder_path, filename), "r", encoding="utf-8") as f:
        json_config = json.loads(f.read())
    return json_config


def base_setup_config() -> Dict[str, Any]:
    """
    Set up the base configuration for a synalinks project.

    Returns:
        (dict): The base configuration for the project.
    """
    questions = []
    config = get_config_json("default_project_config.json")
    default_project_name = config.get("project_name")
    default_project_description = config.get("project_description")
    questions.append(
        inquirer.Text(
            "project_name",
            message=f"What is the project name [{default_project_name}]",
        ),
    )
    questions.append(
        inquirer.Text(
            "project_description",
            message=f"What is the project description [{default_project_description}]",
        ),
    )
    answers = inquirer.prompt(questions)
    questions = []
    if answers:
        for key, answer in answers.items():
            if answer:
                if key == "project_name":
                    config[key] = answer
                    config["package_name"] = to_pkg_name(answer)
                else:
                    config[key] = answer
    else:
        exit(0)
    config["synalinks_version"] = get_version()
    return config


def models_setup_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Set up the language and embedding model configuration for a synalinks project.

    Args:
        config (dict): The existing project configuration.

    Returns:
        (dict): The updated project configuration with language model settings.
    """
    questions = []
    lm_config = get_config_json("language_models.json")
    em_config = get_config_json("embedding_models.json")
    model_providers = list(lm_config.keys())
    questions.append(
        inquirer.List(
            "model_provider",
            message="Select the model provider to use",
            choices=model_providers,
        ),
    )
    answers = inquirer.prompt(questions)
    questions = []
    if answers:
        config = {**config, **answers}
        if answers.get("model_provider") in lm_config:
            questions.append(
                inquirer.List(
                    "language_model",
                    message="Select the language model to use",
                    choices=lm_config.get(answers.get("model_provider")),
                )
            )
        if answers.get("model_provider") in em_config:
            questions.append(
                inquirer.List(
                    "embedding_model",
                    message="Select the embedding model to use",
                    choices=em_config.get(answers.get("model_provider")),
                )
            )
        answers = inquirer.prompt(questions)
        if answers:
            for key, answer in answers.items():
                if key in ("language_model", "embedding_model"):
                    config[key] = answer.split()[0]
                else:
                    config[key] = answer
        else:
            exit(0)
    else:
        exit(0)
    return config


def secrets_setup_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Set up the secrets configuration for a synalinks project.

    Args:
        config (dict): The existing project configuration.

    Returns:
        (dict): The secrets configuration for the project.
    """
    secrets = copy.deepcopy(get_config_json("api_key_config.json"))
    lm_provider = config.get("model_provider", None)
    if lm_provider:
        if lm_provider != "ollama":
            questions = []
            for env_var_name in secrets[lm_provider].keys():
                questions.append(inquirer.Password(env_var_name, message=env_var_name))
            answers = inquirer.prompt(questions)
            if answers:
                for env_var_name, env_var_value in answers.items():
                    if env_var_value:
                        secrets[lm_provider][env_var_name] = env_var_value
    return secrets


def setup_project(config: Dict[str, Any], secrets: Dict[str, Any] = {}):
    """
    Set up a new synalinks project with the given configuration and secrets.

    Args:
        config (dict): The project configuration.
        secrets (dict): The secrets to be stored in the .env file.
    """
    project_dir = os.path.join(os.getcwd(), config["project_name"])
    # Setup project
    queue = deque([(project_dir, PROJECT_STRUCTURE_TEMPLATE)])

    while queue:
        node_dir, node_structure = queue.pop()
        if not os.path.exists(node_dir):
            os.mkdir(node_dir)
        if node_dir == project_dir:
            with open(
                os.path.join(project_dir, PROJECT_CONFIG_FILENAME), "w", encoding="utf-8"
            ) as f:
                f.write(json.dumps(config, indent=2))
        for filename, template in node_structure.items():
            filename = jinja2.Template(filename).render(config=config)
            path = os.path.join(node_dir, filename)
            if isinstance(template, dict):
                queue.append((path, template))
            elif isinstance(template, str):
                with open(path, "w", encoding="utf-8") as f:
                    template = jinja2.Template(template).render(config=config)
                    f.write(template)
    # Setup secrets into an .env file
    model_provider = config.get("model_provider")
    language_model = model_provider + "/" + config.get("language_model")
    embedding_model = model_provider + "/" + config.get("embedding_model")
    variables = [
        f"LANGUAGE_MODEL={language_model}",
        f"EMBEDDING_MODEL={embedding_model}",
    ]
    if model_provider == "ollama":
        variables.append(
            "# To enable your container to communicate with your local ollama "
            "instance, uncomment the following line",
        )
        variables.append("MODEL_API_BASE=http://host.docker.internal:11434")
    else:
        variables.append(
            "# To enable your container to communicate with your local ollama "
            "instance, uncomment the following line",
        )
        variables.append("# MODEL_API_BASE=http://host.docker.internal:11434")
    for lm_provider, env_variables in secrets.items():
        for env_var, env_var_value in env_variables.items():
            variables.append(f"{env_var}={env_var_value}")
    with open(os.path.join(project_dir, ".env"), "w", encoding="utf-8") as f:
        f.write("\n".join(variables))


def maybe_setup_project() -> Dict[str, Any]:
    """
    Set up a new synalinks project if not already inside one.

    Returns:
        (dict): The project configuration.
    """
    if is_inside_synalinks_project():
        config = get_synalinks_project_config()
    else:
        click.echo("Command executed outside a synalinks project.")
        create_new_project = (
            inquirer.confirm(
                "new_project",
                message="Do you want to create a new project",
                default=False,
            ),
        )
        if create_new_project:
            config = base_setup_config()
            config = models_setup_config(config)
            secrets = secrets_setup_config(config)
            setup_project(config, secrets)
        else:
            click.echo("Aborting")
            exit(0)
    return config



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import Embeddings
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.ops.operation import Operation
from synalinks.src.saving import serialization_lib


class Embedding(Operation):
    """Extract the embedding vectors from a data model using an `EmbeddingModel`.

    Args:
        embedding_model (EmbeddingModel): The embedding model to use.
        name (str): Optional. The name of the operation.
        description (str): Optional. Description of the operation.
        **kwargs (keyword warguments): Additional keyword arguments send to the
            embedding model.
    """

    def __init__(
        self,
        embedding_model=None,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.embedding_model = embedding_model
        self.em_kwargs = kwargs

    async def call(self, x):
        texts = tree.flatten(tree.map_structure(lambda field: str(field), x.get_json()))
        embeddings = await self.embedding_model(texts, **self.em_kwargs)
        return JsonDataModel(data_model=Embeddings(**embeddings), name=self.name)

    async def compute_output_spec(self, x):
        return SymbolicDataModel(schema=Embeddings.get_schema(), name=self.name)

    def get_config(self):
        config = {
            "name": self.name,
            "description": self.description,
        }
        embedding_model_config = serialization_lib.serialize_synalinks_object(
            self.embedding_model
        )
        config.update({"em_kwargs": self.em_kwargs})
        return {"embedding_model": embedding_model_config, **config}

    @classmethod
    def from_config(cls, config):
        embedding_model = serialization_lib.deserialize_synalinks_object(
            config.pop("embedding_model")
        )
        em_kwargs = config.pop("em_kwargs")
        return cls(embedding_model=embedding_model, **config, **em_kwargs)


@synalinks_export(["synalinks.ops.embedding", "synalinks.ops.embedding_models.embedding"])
async def embedding(x, embedding_model=None, name=None, description=None, **kwargs):
    """Extract the embedding vectors from a data model using an `EmbeddingModel`.

    Embedding consist in converting the given data_model into a vector representation.
    This function always output a data model that uses `Embeddings` schema.

    If the input data model have multiple fields, each one is embedded.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data_model
        embedding_model (EmbeddingModel): The embedding model to use
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.
        **kwargs (keyword arguments): Additional keywords forwarded to the
            EmbeddingModel call.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data_model
    """
    if embedding_model is None:
        raise ValueError("You should provide the `embedding_model` argument")
    if any_symbolic_data_models(x):
        return await Embedding(
            embedding_model=embedding_model,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await Embedding(
        embedding_model=embedding_model,
        name=name,
        description=description,
        **kwargs,
    )(x)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List
from typing import Literal
from typing import Union

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Entities
from synalinks.src.backend import Entity
from synalinks.src.backend import KnowledgeGraph
from synalinks.src.backend import Relation
from synalinks.src.backend.common.json_data_model import JsonDataModel


class JsonDataModelTest(testing.TestCase):
    def test_init_with_data_model(self):
        class Query(DataModel):
            query: str

        json_data_model = JsonDataModel(
            data_model=Query(query="What is the capital of France?")
        )

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        self.assertEqual(json_data_model.get_json(), expected_json)
        self.assertEqual(json_data_model.get_schema(), expected_schema)

    def test_init_with_data_model_non_instanciated(self):
        class Query(DataModel):
            query: str

        with self.assertRaisesRegex(ValueError, "Couldn't get the JSON data"):
            _ = JsonDataModel(data_model=Query)

    def test_init_with_data_model_non_instanciated_and_value(self):
        class Query(DataModel):
            query: str

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        json_data_model = JsonDataModel(
            data_model=Query,
            json=expected_json,
        )

        self.assertEqual(json_data_model.get_json(), expected_json)
        self.assertEqual(json_data_model.get_schema(), expected_schema)

    def test_init_with_dict(self):
        schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        value = {"query": "What is the capital of France?"}

        json_data_model = JsonDataModel(schema=schema, json=value)

        self.assertEqual(json_data_model.get_json(), value)
        self.assertEqual(json_data_model.get_schema(), schema)

    def test_representation(self):
        class Query(DataModel):
            query: str

        json_data_model = JsonDataModel(
            data_model=Query(query="What is the capital of France?")
        )

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        self.assertEqual(
            str(json_data_model),
            f"<JsonDataModel schema={expected_schema}, json={expected_json}>",
        )

    def test_get_nested(self):
        class Document(Entity):
            label: Literal["Document"]
            title: str
            text: str

        class Chunk(Entity):
            label: Literal["Chunk"]
            text: str

        class IsPartOf(Relation):
            label: Literal["IsPartOf"]
            subj: Chunk
            obj: Document

        doc = Document(
            label="Document",
            title="test",
            text="test document",
        )

        chunk = Chunk(label="Chunk", text="text chunk")

        rel = IsPartOf(
            subj=chunk,
            label="IsPartOf",
            obj=doc,
        )

        subj = rel.to_json_data_model().get_nested_entity("subj")
        obj = rel.to_json_data_model().get_nested_entity("obj")
        self.assertTrue(subj is not None)
        self.assertTrue(subj.get_schema() == Chunk.get_schema())
        self.assertTrue(obj is not None)
        self.assertTrue(obj.get_schema() == Document.get_schema())

    def test_get_nested_union(self):
        class Document(Entity):
            label: Literal["Document"]
            title: str
            text: str

        class Chunk(Entity):
            label: Literal["Chunk"]
            text: str

        class IsPartOf(Relation):
            label: Literal["IsPartOf"]
            subj: Chunk
            obj: Document

        doc = Document(
            label="Document",
            title="test",
            text="test document",
        )

        chunk = Chunk(label="Chunk", text="text chunk")

        rel = IsPartOf(
            subj=chunk,
            label="IsPartOf",
            obj=doc,
        )

        subj = rel.to_json_data_model().get_nested_entity("subj")
        obj = rel.to_json_data_model().get_nested_entity("obj")
        self.assertTrue(subj is not None)
        self.assertTrue(subj.get_schema() == Chunk.get_schema())
        self.assertTrue(obj is not None)
        self.assertTrue(obj.get_schema() == Document.get_schema())

    def test_get_nested_list(self):
        class Document(Entity):
            label: Literal["Document"]
            text: str

        class Documents(Entities):
            entities: List[Document]

        docs = Documents(
            entities=[
                Document(
                    label="Document",
                    text="test document",
                ),
                Document(
                    label="Document",
                    text="another test document",
                ),
            ]
        )

        for doc in docs.to_json_data_model().get_nested_entity_list("entities"):
            self.assertTrue(doc is not None)
            self.assertTrue(doc.get_schema() == Document.get_schema())

    def test_get_nested_list_union(self):
        class Document(Entity):
            label: Literal["Document"]
            text: str

        class Chunk(Entity):
            label: Literal["Chunk"]
            text: str

        class DocumentsAndChunks(Entities):
            entities: List[Union[Document, Chunk]]

        docs = DocumentsAndChunks(
            entities=[
                Document(
                    label="Document",
                    text="test document",
                ),
                Chunk(
                    label="Chunk",
                    text="test chunk",
                ),
            ]
        )

        docs = docs.to_json_data_model().get_nested_entity_list("entities")

        self.assertTrue(docs is not None)
        self.assertTrue(docs[0].get_schema() == Document.get_schema())
        self.assertTrue(docs[1].get_schema() == Chunk.get_schema())

    def test_get_nested_data_model_knowledge_graph(self):
        class Document(Entity):
            label: Literal["Document"]
            text: str

        class Chunk(Entity):
            label: Literal["Chunk"]
            text: str

        class IsPartOf(Relation):
            subj: Union[Chunk, Document]
            label: Literal["IsPartOf"]
            obj: Document

        class IsReferringTo(Relation):
            subj: Document
            label: Literal["IsReferringTo"]
            obj: Document

        class DocumentGraph(KnowledgeGraph):
            entities: List[Union[Document, Chunk]]
            relations: List[Union[IsPartOf, IsReferringTo]]

        doc1 = Document(
            label="Document",
            text="test document",
        )
        doc2 = Document(
            label="Document",
            text="another test document",
        )

        rel1 = IsPartOf(
            subj=doc1,
            label="IsPartOf",
            obj=doc2,
        )

        docs = DocumentGraph(
            entities=[
                doc1,
                doc2,
            ],
            relations=[rel1],
        )

        entities = docs.to_json_data_model().get_nested_entity_list("entities")
        self.assertTrue(entities is not None)
        self.assertTrue(entities[0].get_schema() == Document.get_schema())
        self.assertTrue(entities[1].get_schema() == Document.get_schema())

        relations = docs.to_json_data_model().get_nested_entity_list("relations")
        self.assertTrue(relations is not None)
        self.assertTrue(relations[0].get_schema() == IsPartOf.get_schema())

        subj = relations[0].get_nested_entity("subj")
        obj = relations[0].get_nested_entity("obj")
        self.assertTrue(subj.get_schema() == Document.get_schema())
        self.assertTrue(obj.get_schema() == Document.get_schema())

    def test_contains_json_data_model(self):
        class Foo(DataModel):
            foo: str

        class FooBar(DataModel):
            foo: str
            bar: str

        class Bar(DataModel):
            bar: str

        foo_json = JsonDataModel(data_model=Foo(foo="a"))
        foobar_json = JsonDataModel(data_model=FooBar(foo="a", bar="b"))
        bar_json = JsonDataModel(data_model=Bar(bar="c"))

        self.assertTrue(foo_json in foobar_json)
        self.assertFalse(bar_json in foo_json)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import os
import warnings

from synalinks.src.backend import is_entity
from synalinks.src.backend import is_relation
from synalinks.src.backend import is_similarity_search
from synalinks.src.backend import is_triplet_search
from synalinks.src.knowledge_bases.database_adapters import DatabaseAdapter
from synalinks.src.knowledge_bases.database_adapters.neo4j_adapter import Neo4JAdapter
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.utils.naming import to_snake_case


class MemGraphAdapter(Neo4JAdapter):
    def __init__(
        self,
        uri=None,
        entity_models=None,
        relation_models=None,
        embedding_model=None,
        metric="cosine",
        wipe_on_start=False,
    ):
        self.db_name = os.getenv("MEMGRAPH_DATABASE", "memgraph")
        self.username = os.getenv("MEMGRAPH_USERNAME", "memgraph")
        self.password = os.getenv("MEMGRAPH_PASSWORD", "memgraph")

        DatabaseAdapter.__init__(
            self,
            uri=uri.replace("memgraph://", "bolt://"),
            entity_models=entity_models,
            relation_models=relation_models,
            embedding_model=embedding_model,
            metric=metric,
            wipe_on_start=wipe_on_start,
        )

    def wipe_database(self):
        run_maybe_nested(self.query("MATCH (n) DETACH DELETE n;"))
        result = run_maybe_nested(self.query("SHOW VECTOR INDEXES"))
        for indexes in result:
            index_name = indexes["index_name"]
            query = f"DROP VECTOR INDEX {index_name};"
            run_maybe_nested(self.query(query))

    def create_vector_index(self):
        metric_mapping = {"cosine": "cos", "euclidean": "l2sq"}
        metric = metric_mapping[self.metric]

        for entity_model in self.entity_models:
            try:
                node_label = self.sanitize_label(entity_model.get_schema().get("title"))
                index_name = to_snake_case(node_label)
                query = "\n".join(
                    [
                        f"CREATE VECTOR INDEX {index_name} ",
                        f"ON :{node_label}(embedding)",
                        "WITH CONFIG {",
                        f'  "dimension": {self.embedding_dim}, ',
                        f'  "metric": "{metric}", ',
                        '  "capacity": 1000',
                        "};",
                    ]
                )
                run_maybe_nested(
                    self.query(query),
                )
            except Exception:
                pass

    async def update(
        self,
        data_model,
        threshold=0.8,
    ):
        if is_relation(data_model):
            subj = data_model.get_nested_entity("subj")
            obj = data_model.get_nested_entity("obj")
            relation_label = data_model.get("label")
            subj_label = self.sanitize_label(subj.get("label"))
            subj_vector = subj.get("embedding")
            obj_label = self.sanitize_label(obj.get("label"))
            obj_vector = obj.get("embedding")

            if not subj_vector or not obj_vector:
                warnings.warn(
                    "No embedding found for `subj` or `obj`:"
                    " Entities and relations needs to be embedded. "
                    "Use `Embedding` module before `UpdateKnowledge`. "
                    "Skipping update."
                )
                return

            relation_properties = self.sanitize_properties(data_model.get_json())
            set_clauses = []
            for key in relation_properties.keys():
                if key not in ("subj", "obj"):
                    set_clauses.append(f"r.{key} = ${key}")
            set_statement = "SET " + ", ".join(set_clauses) if set_clauses else ""

            query = "\n".join(
                [
                    "CALL vector_search.search($subjIndexName, 1, $subjVector)",
                    "YIELD node AS s, similarity AS subj_score",
                    "WITH s, subj_score",
                    "WHERE subj_score >= $threshold",
                    "CALL vector_search.search($objIndexName, 1, $objVector)",
                    "YIELD node AS o, similarity AS obj_score",
                    "WITH s, subj_score, o, obj_score",
                    "WHERE obj_score >= $threshold",
                    f"MERGE (s)-[r:{relation_label}]->(o)",
                    (
                        set_statement
                        if set_statement
                        else "// No additional properties to set"
                    ),
                ]
            )
            params = {
                "subjIndexName": to_snake_case(subj_label),
                "objIndexName": to_snake_case(obj_label),
                "threshold": threshold,
                "subjVector": subj_vector,
                "objVector": obj_vector,
                **relation_properties,
            }
            await self.query(query, params=params)
        elif is_entity(data_model):
            node_label = self.sanitize_label(data_model.get("label"))
            vector = data_model.get("embedding")

            if not vector:
                warnings.warn(
                    "Entities need to be embedded. "
                    "Make sure to use `Embedding` module before `UpdateKnowledge`. "
                    "Skipping update."
                )
                return

            node_properties = self.sanitize_properties(data_model.get_json())
            set_clauses = []
            for key in node_properties.keys():
                set_clauses.append(f"n.{key} = ${key}")
            set_statement = "SET " + ", ".join(set_clauses) if set_clauses else ""

            query = "\n".join(
                [
                    "CALL vector_search.search($indexName, 1, $vector)",
                    "YIELD node, similarity AS score",
                    "WITH node, score",
                    "WHERE score >= $threshold",
                    "WITH count(node) AS existing_count",
                    "WHERE existing_count = 0",
                    f"CREATE (n:{node_label})",
                    (
                        set_statement
                        if set_statement
                        else "// No additional properties to set"
                    ),
                ]
            )
            params = {
                "indexName": to_snake_case(node_label),
                "threshold": threshold,
                "vector": vector,
                **node_properties,
            }
            await self.query(query, params=params)
        else:
            raise ValueError(
                "The parameter `data_model` must be an `Entity` or `Relation` instance"
            )

    async def similarity_search(
        self,
        similarity_search,
        k=10,
        threshold=0.7,
    ):
        if not is_similarity_search(similarity_search):
            raise ValueError(
                "The `similarity_search` argument "
                "should be a `SimilaritySearch` data model"
            )
        text = similarity_search.get("similarity_search")
        entity_label = similarity_search.get("entity_label")
        vector = (await self.embedding_model(texts=[text]))["embeddings"][0]

        index_name = to_snake_case(self.sanitize_label(entity_label))

        query = "\n".join(
            [
                "CALL vector_search.search(",
                " $indexName,",
                " $numberOfNearestNeighbours,",
                " $vector) YIELD node AS node, similarity AS score",
                "WITH node, score",
                "WHERE score >= $threshold",
                "RETURN node AS node, score",
                "LIMIT $numberOfNearestNeighbours",
            ]
        )
        params = {
            "indexName": index_name,
            "numberOfNearestNeighbours": k,
            "threshold": threshold,
            "vector": vector,
        }
        result = await self.query(query, params=params)
        return result

    async def triplet_search(
        self,
        triplet_search,
        k=10,
        threshold=0.7,
    ):
        if not is_triplet_search(triplet_search):
            raise ValueError(
                "The `triplet_search` argument should be a `TripletSearch` data model"
            )

        subject_label = triplet_search.get("subject_label")
        subject_label = self.sanitize_label(subject_label)
        subject_similarity_search = triplet_search.get("subject_similarity_search")
        relation_label = triplet_search.get("relation_label")
        relation_label = self.sanitize_label(relation_label)
        object_label = triplet_search.get("object_label")
        object_label = self.sanitize_label(object_label)
        object_similarity_search = triplet_search.get("object_similarity_search")

        params = {
            "numberOfNearestNeighbours": k,
            "threshold": threshold,
        }
        query_lines = []

        has_subject_similarity = (
            subject_similarity_search and subject_similarity_search != "?"
        )
        has_object_similarity = (
            object_similarity_search and object_similarity_search != "?"
        )

        if has_subject_similarity and has_object_similarity:
            subject_vector = (
                await self.embedding_model(texts=[subject_similarity_search])
            )["embeddings"][0]
            object_vector = (
                await self.embedding_model(texts=[object_similarity_search])
            )["embeddings"][0]
            params["subjVector"] = subject_vector
            params["objVector"] = object_vector

            params["subjIndexName"] = to_snake_case(subject_label)
            query_lines.append(
                (
                    "CALL vector_search.search("
                    "$subjIndexName, $numberOfNearestNeighbours, $subjVector)"
                )
            )
            query_lines.extend(
                [
                    "YIELD node AS subj, similarity AS subj_score",
                    "WITH subj, subj_score",
                    "WHERE subj_score >= $threshold",
                ]
            )
            params["objIndexName"] = to_snake_case(object_label)
            query_lines.append(
                (
                    "CALL vector_search.search("
                    "$objIndexName, $numberOfNearestNeighbours, $objVector)"
                )
            )
            query_lines.extend(
                [
                    "YIELD node AS obj, similarity AS obj_score",
                    "WITH subj, subj_score, obj, obj_score",
                    "WHERE obj_score >= $threshold",
                    "WITH subj, subj_score, obj, obj_score",
                ]
            )
            query_lines.append(f"MATCH (subj)-[relation:{relation_label}]->(obj)")
            query_lines.append("WITH subj, subj_score, relation, obj, obj_score")

        elif has_subject_similarity:
            subject_vector = (
                await self.embedding_model(texts=[subject_similarity_search])
            )["embeddings"][0]
            params["subjVector"] = subject_vector

            params["subjIndexName"] = to_snake_case(subject_label)
            query_lines.append(
                (
                    "CALL vector_search.search("
                    "$subjIndexName, $numberOfNearestNeighbours, $subjVector)"
                )
            )
            query_lines.extend(
                [
                    "YIELD node AS subj, similarity AS subj_score",
                    "WITH subj, subj_score",
                    "WHERE subj_score >= $threshold",
                ]
            )
            query_lines.append(f"MATCH (subj)-[relation:{relation_label}]->(obj)")
            query_lines.append("WITH subj, subj_score, relation, obj, 1.0 AS obj_score")

        elif has_object_similarity:
            object_vector = (
                await self.embedding_model(texts=[object_similarity_search])
            )["embeddings"][0]
            params["objVector"] = object_vector
            params["objIndexName"] = to_snake_case(object_label)
            query_lines.append(
                (
                    "CALL vector_search.search("
                    "$objIndexName, $numberOfNearestNeighbours, $objVector)"
                )
            )
            query_lines.extend(
                [
                    "YIELD node AS obj, similarity AS obj_score",
                    "WITH obj, obj_score",
                    "WHERE obj_score >= $threshold",
                ]
            )
            query_lines.append(f"MATCH (subj)-[relation:{relation_label}]->(obj)")
            query_lines.append("WITH subj, 1.0 AS subj_score, relation, obj, obj_score")
        else:
            query_lines.append(
                (
                    f"MATCH (subj:{subject_label})"
                    f"-[relation:{relation_label}]->"
                    f"(obj:{object_label})"
                )
            )
            query_lines.append(
                "WITH subj, 1.0 AS subj_score, relation, obj, 1.0 AS obj_score"
            )
        query_lines.append(
            (
                "RETURN subj, properties(relation) AS relation, obj, "
                "sqrt(subj_score * obj_score) AS score"
            )
        )
        query_lines.append("LIMIT $numberOfNearestNeighbours")
        query = "\n".join(query_lines)
        return await self.query(query, params)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.core.decision import Decision
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib


@synalinks_export(["synalinks.modules.Branch", "synalinks.Branch"])
class Branch(Module):
    """Use a `LanguageModel` to select which module to call based on an arbitrary
        input, a question and a list of labels.

    The selected branch output the data model computed using
    the inputs and module's branch, while the others output `None`.

    Example:

    ```python
    import synalinks
    import asyncio

    async def main():
        class Query(synalinks.DataModel):
            query: str

        class Answer(synalinks.DataModel):
            answer: str

        class AnswerWithCritique(synalinks.DataModel):
            thinking: str
            critique: str
            answer: str

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        x0 = synalinks.Input(data_model=Query)
        (x1, x2) = await synalinks.Branch(
            question="What is the difficulty level of the above query?",
            labels=["easy", "difficult"],
            branches=[
                synalinks.Generator(
                    data_model=Answer,
                    language_model=language_model,
                ),
                synalinks.Generator(
                    data_model=AnswerWithCritique,
                    language_model=language_model,
                ),
            ],
            language_model=language_model,
        )(x0)
        x3 = x1 | x2

        program = synalinks.Program(
            inputs=x0,
            outputs=x3,
            name="adaptative_chain_of_thought",
            description="Useful to answer step by step only when needed",
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Args:
        question (str): The question to ask.
        labels (list): The list of labels to choose from (strings).
        branches (list): The list of modules or programs to select from.
        inject_decision (bool): If True, inject the decision to the branch inputs.
            (default to True).
        return_decision (bool): If True, return the decision with the branch outputs.
            (default to True).
        language_model (LanguageModel): The language model to use.
        prompt_template (str): The default jinja2 prompt template
            to use (see `Generator`).
        examples (list): The default examples to use in the prompt
            (see `Decision`).
        instructions (list): The default instructions to use (see `Decision`).
        seed_instructions (list): Optional. A list of instructions to use as seed for the
            optimization. If not provided, use the default instructions as seed.
        temperature (float): Optional. The temperature for the LM call.
        use_inputs_schema (bool): Optional. Whether or not use the inputs
            schema in the decision prompt (Default to False) (see `Decision`).
        use_outputs_schema (bool): Optional. Whether or not use the outputs
            schema in the decision prompt (Default to False) (see `Decision`).
        decision_type (bool): Optional. The type of decision module to use.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        question=None,
        labels=None,
        branches=None,
        inject_decision=True,
        return_decision=True,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        seed_instructions=None,
        temperature=0.0,
        use_inputs_schema=False,
        use_outputs_schema=False,
        decision_type=Decision,
        name=None,
        description=None,
        trainable=True,
        **kwargs,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        if not branches:
            raise ValueError("The `branches` argument must be provided.")
        if not isinstance(branches, list):
            raise ValueError("The `branches` must be a list of `Module` or `Program`.")
        if len(labels) != len(branches):
            raise ValueError("The `labels` and `branches` must have the same length.")
        self.question = question
        self.labels = labels
        self.branches = {labels[i]: m for i, m in enumerate(branches)}
        self.inject_decision = inject_decision
        self.return_decision = return_decision
        self.language_model = language_model
        self.prompt_template = prompt_template
        self.examples = examples
        self.instructions = instructions
        self.seed_instructions = seed_instructions
        self.temperature = temperature
        self.use_inputs_schema = use_inputs_schema
        self.use_outputs_schema = use_outputs_schema
        self.decision = decision_type(
            question=self.question,
            labels=self.labels,
            language_model=self.language_model,
            prompt_template=self.prompt_template,
            examples=self.examples,
            instructions=self.instructions,
            seed_instructions=self.seed_instructions,
            temperature=self.temperature,
            use_inputs_schema=self.use_inputs_schema,
            use_outputs_schema=self.use_outputs_schema,
            name=self.name + "_decision",
        )

    async def call(self, inputs, training=False):
        outputs = [None] * len(self.branches)

        if not inputs:
            return tuple(outputs)

        decision = await self.decision(
            inputs,
            training=training,
        )
        choice = decision.get("choice", decision.get("choices"))

        if not choice:
            return tuple(outputs)

        if self.inject_decision:
            inputs = await ops.concat(
                inputs,
                decision,
                name=self.name + "_inputs_with_decision",
            )

        tasks = []

        async def execute_branch(
            inputs, module=None, decision=None, return_decision=False
        ):
            if not inputs:
                return None
            if return_decision:
                return await ops.logical_and(
                    decision,
                    await module(inputs),
                )
            else:
                return await module(inputs)

        for label in self.labels:
            module = self.branches[label]
            selected = False
            if isinstance(choice, str):
                if label == choice:
                    selected = True
            elif isinstance(choice, (list, set)):
                if label in choice:
                    selected = True
            if selected and module:
                tasks.append(
                    execute_branch(
                        inputs,
                        module,
                        decision,
                        return_decision=self.return_decision,
                    )
                )
            else:
                tasks.append(execute_branch(None))
        outputs = await asyncio.gather(*tasks)
        return tuple(outputs)

    async def compute_output_spec(self, inputs, training=False):
        outputs = []
        decision = await self.decision(
            inputs,
            training=training,
        )
        if self.inject_decision:
            inputs = await ops.concat(
                inputs,
                decision,
                name=self.name + "_inputs_with_decision",
            )
        for label in self.labels:
            module = self.branches[label]
            if self.return_decision:
                outputs.append(
                    await ops.logical_and(
                        decision,
                        await module(
                            inputs,
                            training=training,
                        ),
                        name=self.name + "_with_decision",
                    )
                )
            else:
                outputs.append(
                    await module(
                        inputs,
                        training=training,
                    )
                )
        return tuple(outputs)

    def get_config(self):
        config = {
            "question": self.question,
            "labels": self.labels,
            "inject_decision": self.inject_decision,
            "return_decision": self.return_decision,
            "prompt_template": self.prompt_template,
            "examples": self.examples,
            "instructions": self.instructions,
            "seed_instructions": self.seed_instructions,
            "temperature": self.temperature,
            "use_inputs_schema": self.use_inputs_schema,
            "use_outputs_schema": self.use_outputs_schema,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model
            )
        }
        branches_config = {
            "branches": [
                serialization_lib.serialize_synalinks_object(branch)
                for branch in self.branches.values()
            ]
        }
        return {**config, **language_model_config, **branches_config}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        branches = [
            serialization_lib.deserialize_synalinks_object(
                branch_config, custom_objects=custom_objects
            )
            for branch_config in config.pop("branches")
        ]
        return cls(language_model=language_model, branches=branches, **config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import numpy as np
from datasets import load_dataset

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field


class MathQuestion(DataModel):
    question: str = Field(
        description="The math word problem",
    )


class NumericalAnswerWithThinking(DataModel):
    thinking: str = Field(
        description="Your step by step thinking",
    )
    answer: float = Field(
        description="The numerical answer",
    )


@synalinks_export("synalinks.datasets.gsm8k.get_input_data_model")
def get_input_data_model():
    """
    Returns GSM8K input data_model for pipeline configurations.

    Returns:
        (DataModel): The GSM8K input data_model
    """
    return MathQuestion


@synalinks_export("synalinks.datasets.gsm8k.get_output_data_model")
def get_output_data_model():
    """
    Returns GSM8K output data_model for pipeline configurations.

    Returns:
        (DataModel): The GSM8K output data_model
    """
    return NumericalAnswerWithThinking


@synalinks_export("synalinks.datasets.gsm8k.load_data")
def load_data():
    """
    Load and format data from HuggingFace

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = synalinks.datasets.gsm8k.load_data()
    ```

    Returns:
        (tuple): The train and test data ready for training
    """
    dataset = load_dataset("gsm8k", "main")

    x_train = []
    y_train = []
    x_test = []
    y_test = []

    for data_point in dataset["train"]:
        question = data_point["question"]
        thinking = data_point["answer"].split("####")[0].strip()
        answer = data_point["answer"].split("####")[-1].strip()
        x_train.append(MathQuestion(question=question))
        y_train.append(
            NumericalAnswerWithThinking(
                thinking=thinking,
                answer=float(answer.replace(",", "")),
            )
        )

    for data_point in dataset["test"]:
        question = data_point["question"]
        thinking = data_point["answer"].split("####")[0].strip()
        answer = data_point["answer"].split("####")[-1].strip()
        x_test.append(MathQuestion(question=question))
        y_test.append(
            NumericalAnswerWithThinking(
                thinking=thinking,
                answer=float(answer.replace(",", "")),
            )
        )

    x_train = np.array(x_train, dtype="object")
    y_train = np.array(y_train, dtype="object")

    x_test = np.array(x_test, dtype="object")
    y_test = np.array(y_test, dtype="object")

    return (x_train, y_train), (x_test, y_test)



# Modified from: keras/src/models/functional_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import JsonDataModel
from synalinks.src.modules import Input
from synalinks.src.programs import Functional
from synalinks.src.programs import Program


class FunctionalTest(testing.TestCase):
    async def test_basic_flow_multi_input(self):
        class Query(DataModel):
            query: str

        class Answer(DataModel):
            answer: str

        input_a = Input(data_model=Query, name="input_a")
        input_b = Input(data_model=Answer, name="input_b")
        outputs = input_a + input_b
        program = Functional([input_a, input_b], outputs, name="basic")

        self.assertEqual(program.name, "basic")
        self.assertIsInstance(program, Functional)
        self.assertIsInstance(program, Program)

        # Eager call
        in_val = [
            JsonDataModel(
                json={"query": "What is the capital of France?"},
                schema=Query.get_schema(),
            ),
            JsonDataModel(json={"answer": "Paris"}, schema=Answer.get_schema()),
        ]
        out_val = await program(in_val)
        expected_schema = {
            "properties": {
                "query": {"title": "Query", "type": "string"},
                "answer": {"title": "Answer", "type": "string"},
            },
            "required": ["query", "answer"],
            "title": "Query",
            "type": "object",
            "additionalProperties": False,
        }
        expected_value = {
            "query": "What is the capital of France?",
            "answer": "Paris",
        }
        self.assertIsInstance(out_val, JsonDataModel)
        self.assertEqual(out_val.get_schema(), expected_schema)
        self.assertEqual(out_val.get_json(), expected_value)

        # Eager call with data_models
        in_val = [
            Query(query="What is the capital of France?"),
            Answer(answer="Paris"),
        ]
        out_val = await program(in_val)
        self.assertIsInstance(out_val, JsonDataModel)
        self.assertEqual(out_val.get_schema(), expected_schema)
        self.assertEqual(out_val.get_json(), expected_value)

        # Symbolic call
        input_a_2 = Input(data_model=Query, name="input_a_2")
        input_b_2 = Input(data_model=Answer, name="input_b_2")
        in_val = [input_a_2, input_b_2]
        out_val = await program(in_val)
        self.assertEqual(out_val.get_schema(), expected_schema)

    async def test_basic_flow_multi_output(self):
        class Query(DataModel):
            query: str

        class Answer(DataModel):
            answer: str

        input_a = Input(data_model=Query, name="input_a")
        input_b = Input(data_model=Answer, name="input_b")
        output_a = input_a
        output_b = input_a + input_b

        program = Functional([input_a, input_b], [output_a, output_b])

        # Eager call
        in_val = [
            JsonDataModel(
                json={"query": "What is the capital of France?"},
                schema=Query.get_schema(),
            ),
            JsonDataModel(json={"answer": "Paris"}, schema=Answer.get_schema()),
        ]
        out_val = await program(in_val)

        expected_schema_a = {
            "properties": {
                "query": {"title": "Query", "type": "string"},
            },
            "required": ["query"],
            "title": "Query",
            "type": "object",
            "additionalProperties": False,
        }

        expected_schema_b = {
            "properties": {
                "query": {"title": "Query", "type": "string"},
                "answer": {"title": "Answer", "type": "string"},
            },
            "required": ["query", "answer"],
            "title": "Query",
            "type": "object",
            "additionalProperties": False,
        }
        self.assertIsInstance(out_val, list)
        self.assertEqual(len(out_val), 2)
        self.assertEqual(out_val[0].get_schema(), expected_schema_a)
        self.assertEqual(out_val[1].get_schema(), expected_schema_b)

        # Eager call with data_models
        in_val = [
            Query(query="What is the capital of France?"),
            Answer(answer="Paris"),
        ]
        out_val = await program(in_val)
        self.assertIsInstance(out_val, list)
        self.assertEqual(len(out_val), 2)
        self.assertEqual(out_val[0].get_schema(), expected_schema_a)
        self.assertEqual(out_val[1].get_schema(), expected_schema_b)

        # Symbolic call
        input_a_2 = Input(data_model=Query, name="input_a_2")
        input_b_2 = Input(data_model=Answer, name="input_b_2")
        out_val = await program([input_a_2, input_b_2])
        self.assertIsInstance(out_val, list)
        self.assertEqual(len(out_val), 2)
        self.assertEqual(out_val[0].get_schema(), expected_schema_a)
        self.assertEqual(out_val[1].get_schema(), expected_schema_b)

    async def test_basic_flow_dict_io(self):
        class Query(DataModel):
            query: str

        class Answer(DataModel):
            answer: str

        input_a = Input(data_model=Query, name="a")
        input_b = Input(data_model=Answer, name="b")

        outputs = input_a + input_b

        with self.assertRaisesRegex(
            ValueError, "All `inputs` values must be SymbolicDataModels"
        ):
            program = Functional({"a": "input_a", "b": input_b}, outputs)

        with self.assertRaisesRegex(
            ValueError, "All `outputs` values must be SymbolicDataModels"
        ):
            program = Functional({"a": input_a, "b": input_b}, "outputs")

        program = Functional({"a": input_a, "b": input_b}, outputs)

        # Eager call
        in_val = {
            "a": Query(query="What is the capital of France?"),
            "b": Answer(answer="Paris"),
        }
        out_val = await program(in_val)
        expected_schema = {
            "properties": {
                "query": {"title": "Query", "type": "string"},
                "answer": {"title": "Answer", "type": "string"},
            },
            "required": ["query", "answer"],
            "title": "Query",
            "type": "object",
            "additionalProperties": False,
        }
        self.assertIsInstance(out_val, JsonDataModel)
        self.assertEqual(out_val.get_schema(), expected_schema)

    async def test_representation(self):
        class Query(DataModel):
            query: str

        class Answer(DataModel):
            answer: str

        input_a = Input(data_model=Query, name="a")
        input_b = Input(data_model=Answer, name="b")
        outputs = input_a + input_b

        program = Functional(
            [input_a, input_b],
            outputs,
            name="concat_basic",
            description="Concatenate two data_models",
        )
        self.assertEqual(
            str(program),
            "<Functional name=concat_basic, "
            "description='Concatenate two data_models', built=True>",
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import ChatMessages
from synalinks.src.backend import is_chat_messages
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.agents.function_calling_agent import FunctionCallingAgent
from synalinks.src.modules.core.input_module import Input
from synalinks.src.programs import Program
from synalinks.src.saving.object_registration import register_synalinks_serializable
from synalinks.src.utils.tool_utils import Tool


@register_synalinks_serializable()
async def calculate(expression: str):
    """Calculate the result of a mathematical expression.

    Args:
        expression (str): The mathematical expression to calculate, such as
            '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
            parentheses, and spaces.
    """
    if not all(char in "0123456789+-*/(). " for char in expression):
        return {
            "result": None,
            "log": "Error: invalid characters in expression",
        }
    try:
        # Evaluate the mathematical expression safely
        result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
        return {
            "result": result,
            "log": "Successfully executed",
        }
    except Exception as e:
        return {
            "result": None,
            "log": f"Error: {e}",
        }


@register_synalinks_serializable()
async def thinking(thinking: str):
    """Think about something.

    Args:
        thinking (str): Your step by step thinking.
    """
    return {
        "thinking": thinking,
    }


@register_synalinks_serializable()
async def get_weather(location: str):
    """Get weather information for a location.
    Args:
        location (str): The location to get weather for.
    """
    # Mock weather data
    weather_data = {
        "New York": {"temp": 22, "condition": "Sunny"},
        "London": {"temp": 15, "condition": "Cloudy"},
        "Tokyo": {"temp": 28, "condition": "Rainy"},
    }

    if location in weather_data:
        return {
            "location": location,
            "temperature": weather_data[location]["temp"],
            "condition": weather_data[location]["condition"],
            "success": True,
        }
    else:
        return {"location": location, "error": "Location not found", "success": False}


@register_synalinks_serializable()
async def failing_tool(should_fail: bool = True):
    """A tool that intentionally fails for testing error handling.
    Args:
        should_fail (bool): Whether the tool should fail.
    """
    if should_fail:
        raise ValueError("This tool was designed to fail")
    return {"status": "success"}


class FunctionCallingAgentTest(testing.TestCase):
    async def test_agent_instantiation(self):
        """Test basic agent instantiation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
        )(inputs)
        program = Program(
            inputs=inputs,
            outputs=outputs,
            name="function_calling_agent_test",
        )
        self.assertIsNotNone(program)

    @patch("litellm.acompletion")
    async def test_autonomous_mode_simple_calculation(self, mock_completion):
        """Test autonomous mode with a simple calculation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=True,
            max_iterations=3,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="autonomous_calculation_test",
        )

        tool_calls = {
            "thinking": "Perform simple arithmetic operation by adding the numbers given in the input.",  # noqa: E501
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "152648 + 485",
                }
            ],
        }

        tool_calls_1 = {
            "thinking": "The user has asked for a simple arithmetic operation, specifically adding 152648 and 485. I have already performed the calculation using the 'calculate' tool and obtained the result as 153133.",  # noqa: E501
            "tool_calls": [],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_1)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 152648 + 485?",
                ),
            ]
        )
        result = await agent(input_messages)

        print("Result:")
        print(result.prettify_json())

        # Verify result structure
        self.assertIsNotNone(result)
        messages = result.get("messages", [])
        self.assertGreater(len(messages), 0)
        self.assertTrue(is_chat_messages(result))

    @patch("litellm.acompletion")
    async def test_autonomous_mode_complex_calculation(self, mock_completion):
        """Test autonomous mode with a more complex calculation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=True,
            max_iterations=5,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="complex_calculation_test",
        )

        tool_calls = {
            "thinking": "First, I will perform the arithmetic operation as instructed. Let's calculate (150 + 250) * 2 / 4. The order of operations is follow BIDMAS/BODMAS which means Brackets, Orders or Powers, Division and Multiplication, Addition and Subtraction. So, first I will add 150 and 250, then multiply the result by 2, divide it by 4 and finally add 100.",  # noqa: E501
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "(150 + 250) * 2 / 4 + 100",
                }
            ],
        }

        tool_calls_1 = {
            "thinking": "The user provided a mathematical expression to calculate. I performed the operation (150 + 250) * 2 / 4 and then added 100 to the result. Now, the result is 300.",  # noqa: E501
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "(150 + 250) * 2 / 4 + 100",
                }
            ],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_1)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content=(
                        "Calculate (150 + 250) * 2 / 4 and then add 100 to the result"
                    ),
                ),
            ]
        )
        result = await agent(input_messages)
        print("Result:")
        print(result.prettify_json())

    # async def test_interactive_mode_single_step(self):
    #     """Test interactive mode with single step execution."""
    #     language_model = LanguageModel(model="ollama/mistral")
    #     tools = [
    #         Tool(calculate),
    #         Tool(thinking),
    #     ]
    #     inputs = Input(data_model=ChatMessages)
    #     outputs = await FunctionCallingAgent(
    #         language_model=language_model,
    #         tools=tools,
    #         autonomous=False,
    #         return_inputs_with_trajectory=True,
    #     )(inputs)
    #     agent = Program(
    #         inputs=inputs,
    #         outputs=outputs,
    #         name="interactive_single_step_test",
    #     )

    #     input_messages = ChatMessages(
    #         messages=[
    #             ChatMessage(
    #                 role="user",
    #                 content="How much is 152648 + 485?",
    #             )
    #         ]
    #     )
    #     result = await agent(input_messages)
    #     print("Interactive Mode Single Step Result:")
    #     print(result.prettify_json())
    #     raise ValueError

    # async def test_interactive_mode_multi_step(self):
    #     """Test interactive mode with multiple steps simulation."""
    #     language_model = LanguageModel(model="ollama/mistral")
    #     tools = [
    #         Tool(calculate),
    #         Tool(thinking),
    #     ]
    #     inputs = Input(data_model=ChatMessages)
    #     outputs = await FunctionCallingAgent(
    #         language_model=language_model,
    #         tools=tools,
    #         autonomous=False,
    #         return_inputs_with_trajectory=True,
    #     )(inputs)
    #     agent = Program(
    #         inputs=inputs,
    #         outputs=outputs,
    #         name="interactive_multi_step_test",
    #     )

    #     input_messages = ChatMessages(
    #         messages=[
    #             ChatMessage(
    #                 role="user",
    #                 content="I need to calculate 100 + 200, then multiply by 3",
    #             )
    #         ]
    #     )

    #     # Simulate multiple interaction steps
    #     max_steps = 3
    #     for step in range(max_steps):
    #         print(f"\n--- Step {step + 1} ---")
    #         result = await agent(input_messages)
    #         print(f"Step {step + 1} Result:")
    #         print(result.prettify_json())

    #         # Get the latest assistant message
    #         messages = result.get("messages", [])
    #         if messages:
    #             last_message = messages[-1]
    #             if last_message.get("role") == "assistant":
    #                 tool_calls = last_message.get("tool_calls", [])
    #                 if not tool_calls:
    #                     print("No more tool calls, stopping interaction")
    #                     break
    #                 # In a real scenario, you would validate/modify tool calls here
    #                 input_messages = result
    #             else:
    #                 break
    #         else:
    #             break

    # async def test_interactive_mode(self):
    #     language_model = LanguageModel(model="ollama/mistral")

    #     tools = [
    #         Tool(calculate),
    #         Tool(thinking),
    #     ]

    #     inputs = Input(data_model=ChatMessages)
    #     outputs = await FunctionCallingAgent(
    #         language_model=language_model,
    #         tools=tools,
    #     )(inputs)

    #     agent = Program(
    #         inputs=inputs,
    #         outputs=outputs,
    #         name="function_calling_agent_test",
    #     )

    #     input_messages = ChatMessages(
    #         messages=[
    #             ChatMessage(
    #                 role="user",
    #                 content="How much is 152648 + 485?",
    #             )
    #         ]
    #     )

    #     result = await agent(input_messages)

    #     print("Result:")
    #     print(result.prettify_json())



# Modified from: keras/src/backend/common/symbolic_scope.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common import global_state


@synalinks_export("synalinks.SymbolicScope")
class SymbolicScope:
    """Scope to indicate the symbolic stage."""

    def __enter__(self):
        self.original_scope = get_symbolic_scope()
        global_state.set_global_attribute("symbolic_scope", self)
        return self

    def __exit__(self, *args, **kwargs):
        global_state.set_global_attribute("symbolic_scope", self.original_scope)


def in_symbolic_scope():
    return global_state.get_global_attribute("symbolic_scope") is not None


def get_symbolic_scope():
    return global_state.get_global_attribute("symbolic_scope")



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import click

from synalinks.cli.banners import banner
from synalinks.cli.utils.setup_utils import base_setup_config
from synalinks.cli.utils.setup_utils import models_setup_config
from synalinks.cli.utils.setup_utils import secrets_setup_config
from synalinks.cli.utils.setup_utils import setup_project
from synalinks.src.version import version as get_version


@click.group()
def magic_cli():
    """Synalinks AI powered command-line interface."""


@magic_cli.command()
def version():
    """Print synalinks version"""
    click.echo(banner())
    click.echo(f"Version: {get_version()}")


@magic_cli.command()
def memory():
    """Synalinks memory"""
    click.echo(banner())


@magic_cli.command()
def init():
    """Initializes a new project"""
    click.echo(banner())
    config = base_setup_config()
    config = models_setup_config(config)
    secrets = secrets_setup_config(config)
    setup_project(config, secrets)


if __name__ == "__main__":
    magic_cli()



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)


from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import dynamic_enum
from synalinks.src.modules.core.generator import Generator
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib


class Question(DataModel):
    question: str = Field(description="The question to ask yourself.")


class DecisionAnswer(DataModel):
    thinking: str = Field(
        description="Your step by step thinking to choose the correct label."
    )
    choice: str = Field(description="The chosen label.")


def default_decision_instructions(labels):
    """The decision default instructions"""
    return f"""
You will be given a question, your task is to answer step-by-step to choose one the following labels: {labels}
""".strip()


@synalinks_export(["synalinks.modules.Decision", "synalinks.Decision"])
class Decision(Module):
    """Perform a decision on the given input based on a question and a list of labels.

    This module dynamically create an `Enum` schema based on the given labels and
    use it to generate a possible answer using structured output.

    This ensure that the LM answer is **always** one of the provided labels.

    Example:

    ```python
    import synalinks
    import asyncio

    async def main():

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        x0 = synalinks.Input(data_model=synalinks.ChatMessages)
        x1 = await synalinks.Decision(
            question="What is the danger level of the discussion?",
            labels=["low", "medium", "high"],
            language_model=language_model,
        )(x0)

        program = synalinks.Program(
            inputs=x0,
            outputs=x1,
            name="discussion_danger_assessment",
            description="This program assesses the level of danger in a discussion.",
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    You can view this module, as performing a single label classification on the input.

    Args:
        question (str): The question to ask.
        labels (list): The list of labels to choose from (strings).
        language_model (LanguageModel): The language model to use.
        prompt_template (str): The default jinja2 prompt template
            to use (see `Generator`).
        examples (list): The default examples to use in the prompt
            (see `Generator`).
        instructions (list): The default instructions to use (see `Generator`).
        seed_instructions (list): Optional. A list of instructions to use as seed for the
            optimization. If not provided, use the default instructions as seed.
        temperature (float): Optional. The temperature for the LM call.
        use_inputs_schema (bool): Optional. Whether or not use the inputs schema in
            the prompt (Default to False) (see `Generator`).
        use_outputs_schema (bool): Optional. Whether or not use the outputs schema in
            the prompt (Default to False) (see `Generator`).
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        question=None,
        labels=None,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        seed_instructions=None,
        temperature=0.0,
        use_inputs_schema=False,
        use_outputs_schema=False,
        name=None,
        description=None,
        trainable=True,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        if not question:
            raise ValueError("The `question` argument must be provided.")
        if not labels:
            raise ValueError("The `labels` argument must be provided.")
        if not isinstance(labels, list):
            raise ValueError("The `labels` parameter must be a list of string.")
        schema = dynamic_enum(DecisionAnswer.get_schema(), "choice", labels)
        self.schema = schema
        self.question = question
        self.labels = labels
        self.language_model = language_model
        self.prompt_template = prompt_template
        self.examples = examples
        if not instructions:
            instructions = default_decision_instructions(self.labels)
        self.instructions = instructions
        self.temperature = temperature
        self.use_inputs_schema = use_inputs_schema
        self.use_outputs_schema = use_outputs_schema
        self.decision = Generator(
            schema=self.schema,
            language_model=self.language_model,
            prompt_template=self.prompt_template,
            examples=self.examples,
            instructions=self.instructions,
            temperature=self.temperature,
            use_inputs_schema=self.use_inputs_schema,
            use_outputs_schema=self.use_outputs_schema,
            name=self.name + "_generator",
        )

    async def call(self, inputs, training=False):
        if not inputs:
            return None
        inputs = await ops.concat(
            inputs,
            Question(question=self.question),
            name=self.name + "_inputs_with_question",
        )
        result = await self.decision(inputs, training=training)
        return result

    def get_config(self):
        config = {
            "question": self.question,
            "labels": self.labels,
            "prompt_template": self.prompt_template,
            "examples": self.examples,
            "instructions": self.instructions,
            "seed_instructions": self.seed_instructions,
            "temperature": self.temperature,
            "use_inputs_schema": self.use_inputs_schema,
            "use_outputs_schema": self.use_outputs_schema,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model
            )
        }
        return {**config, **language_model_config}

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        return cls(language_model=language_model, **config)



# Modified from: keras/src/backend/common/global_state.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import gc
import threading

from synalinks.src.api_export import synalinks_export

GLOBAL_STATE_TRACKER = threading.local()
GLOBAL_SETTINGS_TRACKER = threading.local()


def set_global_attribute(name, value):
    setattr(GLOBAL_STATE_TRACKER, name, value)


def get_global_attribute(name, default=None, set_to_default=False):
    attr = getattr(GLOBAL_STATE_TRACKER, name, None)
    if attr is None and default is not None:
        attr = default
        if set_to_default:
            set_global_attribute(name, attr)
    return attr


@synalinks_export(
    [
        "synalinks.utils.clear_session",
        "synalinks.backend.clear_session",
        "synalinks.clear_session",
    ]
)
def clear_session(free_memory=True):
    """Resets all state generated by synalinks.

    synalinks manages a global state, which it uses to implement the Functional
    model-building API and to uniquify autogenerated modules names.

    If you are creating many models in a loop, this global state will consume
    an increasing amount of memory over time, and you may want to clear it.
    Calling `clear_session()` releases the global state: this helps avoid
    clutter from old programs and modules, especially when memory is limited.

    Args:
        free_memory (bool): Whether to call Python garbage collection.
            It's usually a good practice to call it to make sure
            memory used by deleted objects is immediately freed.
            However, it may take a few seconds to execute, so
            when using `clear_session()` in a short loop,
            you may want to skip it.
    """
    global GLOBAL_STATE_TRACKER
    global GLOBAL_SETTINGS_TRACKER

    GLOBAL_STATE_TRACKER = threading.local()
    GLOBAL_SETTINGS_TRACKER = threading.local()

    if free_memory:
        # Manually trigger garbage collection.
        gc.collect()



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export

# Unique source of truth for the version number.
__version__ = "0.5.200"


@synalinks_export("synalinks.version")
def version():
    return __version__



# Modified from: keras/src/backend/config.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
import logging
import os
import random

import numpy as np

from synalinks.src.api_export import synalinks_export

# The type of float to use throughout a session.
_FLOATX = "float32"

# Epsilon fuzz factor used throughout the codebase.
_EPSILON = 1e-7

# The Synalinks API key.
_SYNALINKS_API_KEY = None

# Default backend: Pydantic.
_BACKEND = "pydantic"

# Default Synalinks api base
_SYNALINKS_API_BASE = "http://localhost:4248"

# Enable monitoring
_ENABLE_OBSERVABILITY = False

# Available backends
_AVAILABLE_BACKEND = ["pydantic"]

# Random seed for reproductability
_RANDOM_SEED = 123

np.random.seed(_RANDOM_SEED)
random.seed(_RANDOM_SEED)


@synalinks_export(["synalinks.config.floatx", "synalinks.backend.floatx"])
def floatx():
    """Return the default float type, as a string.

    E.g. `'bfloat16'`, `'float16'`, `'float32'`, `'float64'`.

    Returns:
        (str): The current default float type.

    Example:

    ```python
    >>> synalinks.config.floatx()
    'float32'
    ```

    """
    return _FLOATX


@synalinks_export(["synalinks.config.set_floatx", "synalinks.backend.set_floatx"])
def set_floatx(value):
    """Set the default float dtype.

    Note: It is not recommended to set this to `"float16"`,
    as this will likely cause numeric stability issues.
    Instead, use `float64` or `float32`.

    Args:
        value (str): The float type between `'bfloat16'`, `'float16'`, `'float32'`,
            or `'float64'`.

    Examples:

    ```python
    >>> synalinks.config.floatx()
    'float32'
    ```

    ```python
    >>> synalinks.config.set_floatx('float64')
    >>> synalinks.config.floatx()
    'float64'
    ```

    ```python
    >>> # Set it back to float32
    >>> synalinks.config.set_floatx('float32')
    ```

    Raises:
        ValueError: In case of invalid value.
    """
    global _FLOATX
    accepted_dtypes = {"bfloat16", "float16", "float32", "float64"}
    if value not in accepted_dtypes:
        raise ValueError(
            f"Unknown `floatx` value: {value}. Expected one of {accepted_dtypes}"
        )
    _FLOATX = str(value)


@synalinks_export(["synalinks.config.epsilon", "synalinks.backend.epsilon"])
def epsilon():
    """Return the value of the fuzz factor used in numeric expressions.

    Returns:
        (float): The epsilon value.

    Example:

    ```python
    >>> synalinks.config.epsilon()
    1e-07
    ```

    """
    return _EPSILON


@synalinks_export(["synalinks.config.set_epsilon", "synalinks.backend.set_epsilon"])
def set_epsilon(value):
    """Set the value of the fuzz factor used in numeric expressions.

    Args:
        value (float): The new value of epsilon.

    Examples:

    ```python
    >>> synalinks.config.epsilon()
    1e-07
    ```

    ```python
    >>> synalinks.config.set_epsilon(1e-5)
    >>> synalinks.config.epsilon()
    1e-05
    ```

    ```python
    >>> # Set it back to the default value.
    >>> synalinks.config.set_epsilon(1e-7)
    ```

    """
    global _EPSILON
    _EPSILON = value


@synalinks_export(
    [
        "synalinks.config.set_seed",
        "synalinks.backend.set_seed",
        "synalinks.set_seed",
    ]
)
def set_seed(value):
    """Set the value of the random seed for reproductability.

    Args:
        value (float): The new value of epsilon.
    """
    global _RANDOM_SEED
    _RANDOM_SEED = value
    np.random.seed(_RANDOM_SEED)
    random.seed(_RANDOM_SEED)


@synalinks_export(
    [
        "synalinks.config.get_seed",
        "synalinks.backend.get_seed",
        "synalinks.get_seed",
    ]
)
def get_seed():
    return _RANDOM_SEED


@synalinks_export(
    [
        "synalinks.config.enable_logging",
        "synalinks.backend.enable_logging",
        "synalinks.enable_logging",
    ]
)
def enable_logging(filename=None, debug=False):
    """
    Configures and enables logging for the application.

    This function sets up the logging configuration for the application, allowing logs to be output
    either to a specified file or to the console. The logging level can be set to DEBUG for more
    verbose logging or INFO for standard logging.

    Args:
        filename (str): The name of the file where logs should be written. If not provided,
            logs will be output to the console.
        debug (bool): If True, sets the logging level to DEBUG. If False, the logging level
            is set to INFO. Defaults to False.

    The log message format includes the timestamp, log level, and the log message itself.
    """
    level = logging.DEBUG if debug else logging.INFO
    if filename:
        logging.basicConfig(
            filename=filename,
            level=level,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )
    else:
        logging.basicConfig(
            level=level,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )


@synalinks_export(
    [
        "synalinks.config.is_observability_enabled",
        "synalinks.backend.is_observability_enabled",
        "synalinks.is_observability_enabled",
    ]
)
def is_observability_enabled():
    """Check if the observability is enabled

    Returns:
        (bool): True if the observability is enabled.
    """
    return _ENABLE_OBSERVABILITY


@synalinks_export(
    [
        "synalinks.config.api_base",
        "synalinks.backend.api_base",
        "synalinks.api_base",
    ]
)
def api_base():
    """Returns the Synalinks api base

    Returns:
        (str): The observability api base
    """
    return _API_BASE


@synalinks_export(
    [
        "synalinks.config.set_api_base",
        "synalinks.backend.set_api_base",
        "synalinks.set_api_base",
    ]
)
def set_api_base(api_base):
    """Set the observability api base

    Args:
        api_base (str): The observability api base
    """
    global _SYNALINKS_API_BASE
    _SYNALINKS_API_BASE = api_base


@synalinks_export(
    [
        "synalinks.config.enable_observability",
        "synalinks.backend.enable_observability",
        "synalinks.enable_observability",
    ]
)
def enable_observability(api_base=None):
    """
    Configures and enables observability for the application.

    This function sets up the observability configuration for the application,
    allowing traces and logs to be send to Synalinks Lab.

    Args:
        api_base (str): Optional. The api base to send the traces and logs to.
    """
    global _SYNALINKS_API_BASE
    global _ENABLE_OBSERVABILITY
    if api_base:
        _SYNALINKS_API_BASE = api_base
    _ENABLE_OBSERVABILITY = True


@synalinks_export(
    [
        "synalinks.config.api_key",
        "synalinks.backend.api_key",
    ]
)
def api_key():
    """Synalinks API key.

    Returns:
        (str): Synalinks API key.

    ```python
    >>> synalinks.config.api_key()
    'my-secret-api-key'
    ```

    """
    return _SYNALINKS_API_KEY


@synalinks_export(
    [
        "synalinks.config.set_api_key",
        "synalinks.backend.set_api_key",
    ]
)
def set_api_key(key):
    """Set Synalinks API key.

    Args:
        key (str): The API key value.

    The API key is retrieved from the env variables at start.

    ```python
    >>> os.environ["SYNALINKS_API_KEY"] = 'my-secret-api-key'
    ```

    Or you can setup it using the config

    ```python
    >>> synalinks.config.set_api_key('my-secret-api-key')
    >>> synalinks.config.api_key()
    'my-secret-api-key'
    ```

    Args:
        key (str): Synalinks API key.
    """
    global _SYNALINKS_API_KEY
    _SYNALINKS_API_KEY = key


# Set synalinks base dir path given synalinks_HOME env variable, if applicable.
# Otherwise either ~/.synalinks or /tmp.
if "SYNALINKS_HOME" in os.environ:
    _synalinks_DIR = os.environ.get("SYNALINKS_HOME")
else:
    _synalinks_base_dir = os.path.expanduser("~")
    if not os.access(_synalinks_base_dir, os.W_OK):
        _synalinks_base_dir = "/tmp"
    _synalinks_DIR = os.path.join(_synalinks_base_dir, ".synalinks")


@synalinks_export(["synalinks.config.synalinks_home", "synalinks.synalinks_home"])
def synalinks_home():
    # Private accessor for the synalinks home location.
    return _synalinks_DIR


@synalinks_export(["synalinks.config.backend", "synalinks.backend.backend"])
def backend():
    """Publicly accessible method for determining the current backend.

    Returns:
        (str): The name of the backend synalinks is currently using. like
            `"pydantic"`.

    Example:

    ```python
    >>> synalinks.config.backend()
    'pydantic'
    ```
    """
    return _BACKEND


@synalinks_export(["synalinks.config.set_backend", "synalinks.backend.set_backend"])
def set_backend(backend):
    if backend not in _AVAILABLE_BACKEND:
        raise ValueError(
            f"Cannot set backend to {backend}, "
            f"you should choose between: {_AVAILABLE_BACKEND}"
        )
    global _BACKEND
    _BACKEND = backend


# Attemp to get the API key from env variables
_api_key = os.getenv("SYNALINKS_API_KEY", None)
if _api_key:
    assert isinstance(_api_key, str)
    set_api_key(_api_key)

# Attempt to read synalinks config file.
_config_path = os.path.expanduser(os.path.join(_synalinks_DIR, "synalinks.json"))
if os.path.exists(_config_path):
    try:
        with open(_config_path) as f:
            _config = json.load(f)
    except ValueError:
        _config = {}
    _backend = _config.get("backend", _BACKEND)
    assert _backend in _AVAILABLE_BACKEND
    _floatx = _config.get("floatx", _FLOATX)
    assert _floatx in {"float16", "float32", "float64"}
    _epsilon = _config.get("epsilon", _EPSILON)
    assert isinstance(_epsilon, float)
    _seed = _config.get("seed", _RANDOM_SEED)
    assert isinstance(_seed, int)
    _api_base = _config.get("api_base", _SYNALINKS_API_BASE)
    assert isinstance(_api_base, str)

    set_backend(_backend)
    set_floatx(_floatx)
    set_epsilon(_epsilon)
    set_seed(_seed)
    set_api_base(_api_base)

# Save config file, if possible.
if not os.path.exists(_synalinks_DIR):
    try:
        os.makedirs(_synalinks_DIR)
    except OSError:
        # Except permission denied and potential race conditions
        # in multi-threaded environments.
        pass

if not os.path.exists(_config_path):
    _config = {
        "backend": _BACKEND,
        "floatx": _FLOATX,
        "epsilon": _EPSILON,
        "seed": _RANDOM_SEED,
        "api_base": _SYNALINKS_API_BASE,
    }
    try:
        with open(_config_path, "w") as f:
            f.write(json.dumps(_config, indent=4))
    except IOError:
        # Except permission denied.
        pass



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.backend import concatenate_json
from synalinks.src.backend import concatenate_schema
from synalinks.src.backend import factorize_json
from synalinks.src.backend import factorize_schema
from synalinks.src.backend import in_mask_json
from synalinks.src.backend import in_mask_schema
from synalinks.src.backend import out_mask_json
from synalinks.src.backend import out_mask_schema
from synalinks.src.backend import prefix_json
from synalinks.src.backend import prefix_schema
from synalinks.src.backend import suffix_json
from synalinks.src.backend import suffix_schema
from synalinks.src.ops.operation import Operation


class Concat(Operation):
    """Concatenate two data models together."""

    def __init__(
        self,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )

    async def call(self, x1, x2):
        if not x1:
            raise ValueError(f"Received x1={x1} and x2={x2}")
        if not x2:
            raise ValueError(f"Received x1={x1} and x2={x2}")
        json = concatenate_json(x1.get_json(), x2.get_json())
        schema = concatenate_schema(x1.get_schema(), x2.get_schema())
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x1, x2):
        schema = concatenate_schema(x1.get_schema(), x2.get_schema())
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(
    [
        "synalinks.ops.concat",
        "synalinks.ops.concatenate",
        "synalinks.ops.json.concat",
        "synalinks.ops.json.concatenate",
    ]
)
async def concat(x1, x2, name=None, description=None):
    """Concatenate two data models together.

    Concatenation consist in creating a new data model containing
    all the elements of the two inputs into a new one.
    Each field name is made unique if needed by adding a numerical suffix `_n`,
    with `n` being an incremental integer.

    This operation is implemented in the `+` Python operator.

    If any of the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    If any of the inputs is None, then an exception is raised.

    If the keys are used more than once, a numerical suffix is added.

    Table:

    | `x1`   | `x2`   | Concat (`+`)      |
    | ------ | ------ | ----------------- |
    | `x1`   | `x2`   | `x1 + x2`         |
    | `x1`   | `None` | `Exception`       |
    | `None` | `x2`   | `Exception`       |
    | `None` | `None` | `Exception`       |

    Args:
        x1 (JsonDataModel | SymbolicDataModel): the first input data model.
        x2 (JsonDataModel | SymbolicDataModel): the second input data model.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model
    """
    if any_symbolic_data_models(x1, x2):
        return await Concat(
            name=name,
            description=description,
        ).symbolic_call(x1, x2)
    return await Concat(
        name=name,
        description=description,
    )(x1, x2)


class And(Operation):
    """Perform a logical `And` operation between data models."""

    def __init__(
        self,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )

    async def call(self, x1, x2):
        if x1 and x2:
            json = concatenate_json(x1.get_json(), x2.get_json())
            schema = concatenate_schema(x1.get_schema(), x2.get_schema())
            return JsonDataModel(json=json, schema=schema, name=self.name)
        elif x1 and not x2:
            return None
        elif not x1 and x2:
            return None
        else:
            return None

    async def compute_output_spec(self, x1, x2):
        schema = concatenate_schema(x1.get_schema(), x2.get_schema())
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.logical_and", "synalinks.ops.json.logical_and"])
async def logical_and(x1, x2, name=None, description=None):
    """Perform a logical `And` operation between two data models.

    If one of the inputs is `None`, then this operation output `None`.
    If both inputs are provided, the output is a concatenation
    of the two given data models.

    This operation is implemented in the Python `&` operator.

    If any of the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    Table:

    | `x1`   | `x2`   | Logical And (`&`) |
    | ------ | ------ | ----------------- |
    | `x1`   | `x2`   | `x1 + x2`         |
    | `x1`   | `None` | `None`            |
    | `None` | `x2`   | `None`            |
    | `None` | `None` | `None`            |

    Args:
        x1 (JsonDataModel | SymbolicDataModel): The first input data model.
        x2 (JsonDataModel | SymbolicDataModel): The second input data model.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel | None): The resulting data model or
            None if the condition is not met.
    """
    if any_symbolic_data_models(x1, x2):
        return await And(
            name=name,
            description=description,
        ).symbolic_call(x1, x2)
    return await And(
        name=name,
        description=description,
    )(x1, x2)


class Or(Operation):
    """Perform a logical `Or` operation between data models."""

    def __init__(
        self,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )

    async def call(self, x1, x2):
        if x1 and x2:
            json = concatenate_json(x1.get_json(), x2.get_json())
            schema = concatenate_schema(x1.get_schema(), x2.get_schema())
            return JsonDataModel(json=json, schema=schema, name=self.name)
        elif x1 and not x2:
            return JsonDataModel(
                json=x1.get_json(), schema=x1.get_schema(), name=self.name
            )
        elif not x1 and x2:
            return JsonDataModel(
                json=x2.get_json(), schema=x2.get_schema(), name=self.name
            )
        else:
            return None

    async def compute_output_spec(self, x1, x2):
        return SymbolicDataModel(schema=x1.get_schema(), name=self.name)


@synalinks_export(["synalinks.ops.logical_or", "synalinks.ops.json.logical_or"])
async def logical_or(x1, x2, name=None, description=None):
    """Perform a logical `Or` between two data models.

    If one of the input is `None`, then output the other one.
    If both inputs are provided, the output is a concatenation
    of the two given data models.

    If any of the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    This operation is implemented in the Python `|` operator.

    Table:

    | `x1`   | `x2`   | Logical Or (`|`) |
    | ------ | ------ | ---------------- |
    | `x1`   | `x2`   | `x1 + x2`        |
    | `x1`   | `None` | `x1`             |
    | `None` | `x2`   | `x2`             |
    | `None` | `None` | `None`           |

    Args:
        x1 (JsonDataModel | SymbolicDataModel): The first input data model.
        x2 (JsonDataModel | SymbolicDataModel): The second input data model.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel | None): The resulting data model or
            None if the condition is not met.
    """
    if any_symbolic_data_models(x1, x2):
        return await Or(
            name=name,
            description=description,
        ).symbolic_call(x1, x2)
    return await Or(
        name=name,
        description=description,
    )(x1, x2)


class Xor(Operation):
    """Perform a logical `Xor` operation between data models."""

    def __init__(
        self,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )

    async def call(self, x1, x2):
        if x1 and x2:
            return None
        elif x1 and not x2:
            return JsonDataModel(
                json=x1.get_json(), schema=x1.get_schema(), name=self.name
            )
        elif not x1 and x2:
            return JsonDataModel(
                json=x2.get_json(), schema=x2.get_schema(), name=self.name
            )
        else:
            return None

    async def compute_output_spec(self, x1, x2):
        return SymbolicDataModel(schema=x1.get_schema(), name=self.name)


@synalinks_export(["synalinks.ops.logical_xor", "synalinks.ops.json.logical_xor"])
async def logical_xor(x1, x2, name=None, description=None):
    """Perform a logical `Xor` between two data models.

    If one of the input is `None`, then output the other one.
    If both inputs are provided, the output is `None`.

    If any of the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    This operation is implemented in the Python `^` operator.

    Table:

    | `x1`   | `x2`   | Logical Xor (`^`)|
    | ------ | ------ | ---------------- |
    | `x1`   | `x2`   | `None`           |
    | `x1`   | `None` | `x1`             |
    | `None` | `x2`   | `x2`             |
    | `None` | `None` | `None`           |

    Args:
        x1 (JsonDataModel | SymbolicDataModel): The first input data model.
        x2 (JsonDataModel | SymbolicDataModel): The second input data model.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel | None): The resulting data model or
            None if the condition is not met.
    """
    if any_symbolic_data_models(x1, x2):
        return await Xor(
            name=name,
            description=description,
        ).symbolic_call(x1, x2)
    return await Xor(
        name=name,
        description=description,
    )(x1, x2)


class Factorize(Operation):
    """Factorize a data model by grouping similar properties together."""

    def __init__(
        self,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )

    async def call(self, x):
        if not x:
            return None
        json = factorize_json(x.get_json())
        schema = factorize_schema(x.get_schema())
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x):
        schema = factorize_schema(x.get_schema())
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.factorize", "synalinks.ops.json.factorize"])
async def factorize(x, name=None, description=None):
    """Factorize a data model by grouping similar properties together.

    Factorization consist in grouping the same properties into lists.
    The property key of the resulting grouped property is changed to its plural form.
    For example `action` become `actions`, or `query` become `queries`.

    If the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    This operation is implemented in `.factorize()`

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model.
    """
    if any_symbolic_data_models(x):
        return await Factorize(
            name=name,
            description=description,
        ).symbolic_call(x)
    return await Factorize(
        name=name,
        description=description,
    )(x)


class OutMask(Operation):
    """Mask specific fields of a data model."""

    def __init__(
        self,
        mask=None,
        recursive=True,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.mask = mask
        self.recursive = recursive

    async def call(self, x):
        if not x:
            return None
        json = out_mask_json(x.get_json(), mask=self.mask, recursive=self.recursive)
        schema = out_mask_schema(x.get_schema(), mask=self.mask, recursive=self.recursive)
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x):
        schema = out_mask_schema(x.get_schema(), mask=self.mask, recursive=self.recursive)
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.out_mask", "synalinks.ops.json.out_mask"])
async def out_mask(x, mask=None, recursive=True, name=None, description=None):
    """Mask specific fields of a data model.

    Out masking consist in removing the properties that match with the keys given
    in the mask. The masking process ignore the numerical suffixes that could be added
    by other operations.

    If the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model.
        mask (list): the input mask (list of keys).
        recursive (bool): Whether or not to remove
            recursively for nested objects (default True).
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model.
    """
    if mask is None:
        raise ValueError("You should specify the `mask` argument")
    if any_symbolic_data_models(x):
        return await OutMask(
            mask=mask,
            recursive=recursive,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await OutMask(
        mask=mask,
        recursive=recursive,
        name=name,
        description=description,
    )(x)


class InMask(Operation):
    """Keep specific fields of a data model."""

    def __init__(
        self,
        mask=None,
        recursive=True,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.mask = mask
        self.recursive = recursive

    async def call(self, x):
        if not x:
            return None
        json = in_mask_json(x.get_json(), mask=self.mask, recursive=self.recursive)
        schema = in_mask_schema(x.get_schema(), mask=self.mask, recursive=self.recursive)
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x):
        schema = in_mask_schema(x.get_schema(), mask=self.mask, recursive=self.recursive)
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.in_mask", "synalinks.ops.json.in_mask"])
async def in_mask(x, mask=None, recursive=True, name=None, description=None):
    """Keep specific fields of a data model.

    In masking consists in keeping the properties that match with the keys given
    in the mask. The masking process ignores the numerical suffixes that could be added
    by other operations.

    If the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model
        mask (list): the input mask (list of keys)
        recursive (bool): Whether or not to keep
            recursively for nested objects (default True).
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model.
    """
    if mask is None:
        raise ValueError("You should specify the `mask` argument")
    if any_symbolic_data_models(x):
        return await InMask(
            mask=mask,
            recursive=recursive,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await InMask(
        mask=mask,
        recursive=recursive,
        name=name,
        description=description,
    )(x)


class Prefix(Operation):
    """Add a prefix to **all** the data model fields (non-recursive)."""

    def __init__(
        self,
        prefix=None,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.prefix = prefix

    async def call(self, x):
        if not x:
            return None
        json = prefix_json(x.get_json(), self.prefix)
        schema = prefix_schema(x.get_schema(), self.prefix)
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x):
        schema = prefix_schema(x.get_schema(), self.prefix)
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.prefix", "synalinks.ops.json.prefix"])
async def prefix(x, prefix=None, name=None, description=None):
    """Add a prefix to **all** the data model fields (non-recursive).

    If the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model
        prefix (str): the prefix to add.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model.
    """
    if prefix is None:
        raise ValueError("You should specify the `prefix` argument")
    if any_symbolic_data_models(x):
        return await Prefix(
            prefix=prefix,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await Prefix(
        prefix=prefix,
        name=name,
        description=description,
    )(x)


class Suffix(Operation):
    """Add a suffix to **all** the data model fields (non-recursive)."""

    def __init__(
        self,
        suffix=None,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.suffix = suffix

    async def call(self, x):
        if not x:
            return None
        json = suffix_json(x.get_json(), self.suffix)
        schema = suffix_schema(x.get_schema(), self.suffix)
        return JsonDataModel(json=json, schema=schema, name=self.name)

    async def compute_output_spec(self, x):
        schema = suffix_schema(x.get_schema(), self.suffix)
        return SymbolicDataModel(schema=schema, name=self.name)


@synalinks_export(["synalinks.ops.suffix", "synalinks.ops.json.suffix"])
async def suffix(x, suffix=None, name=None, description=None):
    """Add a suffix to **all** the data model fields (non-recursive).

    If the data models used is a metaclass or symbolic data model
    the output is a symbolic data model.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model
        suffix (str): the suffix to add.
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model
    """
    if suffix is None:
        raise ValueError("You should specify the `suffix` argument")
    if any_symbolic_data_models(x):
        return await Suffix(
            suffix=suffix,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await Suffix(
        suffix=suffix,
        name=name,
        description=description,
    )(x)



# Modified from: keras/src/backend/common/global_state_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.backend.common import global_state
from synalinks.src.testing import test_case
from synalinks.src.utils.naming import auto_name


class GlobalStateTest(test_case.TestCase):
    def test_clear_session(self):
        name0 = auto_name("somename")
        self.assertEqual(name0, "somename")
        name1 = auto_name("somename")
        self.assertEqual(name1, "somename_1")
        global_state.clear_session()
        name0 = auto_name("somename")
        self.assertEqual(name0, "somename")



# Modified from: keras/src/trainers/trainer.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import inspect
import warnings

import numpy as np

from synalinks.src import backend
from synalinks.src import callbacks as callbacks_module
from synalinks.src import metrics as metrics_module
from synalinks.src import optimizers as optimizers_module
from synalinks.src.backend.common import numpy
from synalinks.src.saving import serialization_lib
from synalinks.src.trainers.compile_utils import CompileMetrics
from synalinks.src.trainers.compile_utils import CompileReward
from synalinks.src.trainers.data_adapters import array_slicing
from synalinks.src.trainers.data_adapters import data_adapter_utils
from synalinks.src.trainers.epoch_iterator import EpochIterator
from synalinks.src.utils import python_utils
from synalinks.src.utils import tracking
from synalinks.src.utils.async_utils import run_maybe_nested


class Trainer:
    def __init__(self):
        self._lock = False
        self._run_eagerly = False
        self.compiled = False
        self.reward = None
        self.steps_per_execution = 1
        # Can be set by callbacks in on_train_begin
        self._initial_epoch = None
        self._compute_reward_has_training_arg = (
            "training" in inspect.signature(self.compute_reward).parameters
        )
        # Placeholders used in `compile`
        self._optimizer = None
        self._compile_reward = None
        self._compile_metrics = None
        self._reward_tracker = None

    @tracking.no_automatic_dependency_tracking
    def compile(
        self,
        optimizer=None,
        reward=None,
        reward_weights=None,
        metrics=None,
        run_eagerly=False,
        steps_per_execution=1,
    ):
        """Configures the program for training.

        Example:

        ```python
        program.compile(
            optimizer=synalinks.optimizers.RandomFewShot(),
            reward=synalinks.rewards.ExactMatch(),
            metrics=[
                synalinks.metrics.MeanMetricWrapper(synalinks.rewards.exact_match),
            ],
        )
        ```

        Args:
            optimizer (Optimizer): Optimizer instance. See `synalinks.optimizers`.
            reward (Reward): Reward function. A `synalinks.rewards.Reward`
                instance. See `synalinks.rewards`. A reward function is
                any callable with the signature `reward = fn(y_true, y_pred)`,
                where `y_true` are the ground truth values, and `y_pred`
                are the program's predictions.
                `y_true` should be a list of batch size length `[d0, .. dN]`.
                `y_pred` should be a list of batch size length `[d0, .. dN]`.
                The reward function should return a float.
            reward_weights (list): Optional list specifying scalar coefficients
                (Python floats) to weight the reward contributions of
                different program outputs. The reward value that will be maximized
                by the program will then be the *weighted sum* of all individual
                rewards, weighted by the `reward_weights` coefficients. It is
                expected to have a 1:1 mapping to the program's outputs.
            metrics (list): List of metrics to be evaluated by the program during
                training and testing. Each of it is a `synalinks.metrics.Metric`
                instance. See `synalinks.metrics`. A function is any callable with the
                signature `result = fn(y_true, y_pred)`.
            run_eagerly (bool): If `True`, this program's forward pass
                will never be compiled. It is recommended to leave this
                as `False` when training (for best performance),
                and to set it to `True` when debugging.
            steps_per_execution (int): The number of batches to run
                during each a single compiled function call. Running multiple
                batches inside a single compiled function call can
                greatly improve performance on TPUs or small programs with a large
                Python overhead. At most, one full epoch will be run each
                execution. If a number larger than the size of the epoch is
                passed, the execution will be truncated to the size of the
                epoch. Note that if `steps_per_execution` is set to `N`,
                `Callback.on_batch_begin` and `Callback.on_batch_end` methods
                will only be called every `N` batches (i.e. before/after
                each compiled function execution).
        """
        self._clear_previous_trainer_metrics()
        self._optimizer = optimizer
        self._optimizer.set_program(self)

        if hasattr(self, "output_names"):
            output_names = self.output_names
        else:
            output_names = None
        if reward is not None:
            self._compile_reward = CompileReward(
                reward, reward_weights, output_names=output_names
            )
            self.reward = reward
        if metrics is not None:
            self._compile_metrics = CompileMetrics(metrics, output_names=output_names)
        self.run_eagerly = run_eagerly
        self.stop_training = False
        self.compiled = True
        self._reward_tracker = metrics_module.Mean(name="reward")
        self.steps_per_execution = steps_per_execution

        self._compile_config = serialization_lib.SerializableDict(
            optimizer=optimizer,
            reward=reward,
            reward_weights=reward_weights,
            metrics=metrics,
            run_eagerly=run_eagerly,
            steps_per_execution=steps_per_execution,
        )

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def metrics(self):
        # Order: reward tracker, individual reward trackers, compiled metrics,
        # custom metrcis, submodule metrics.
        metrics = []
        if self.compiled:
            if self._reward_tracker is not None:
                metrics.append(self._reward_tracker)
            if self._compile_metrics is not None:
                metrics.append(self._compile_metrics)
            if self._compile_reward is not None:
                metrics.extend(self._compile_reward.metrics)
        metrics.extend(self._metrics)
        for module in self._flatten_modules(include_self=False):
            if isinstance(module, Trainer):
                # All Trainer-related metrics in submodules should be ignored
                # because a new Trainer has been instantiated.
                continue
            metrics.extend(module.metrics)
        return metrics

    @property
    def metrics_names(self):
        return [m.name for m in self.metrics]

    def reset_metrics(self):
        for m in self.metrics:
            m.reset_state()

    def _get_own_metrics(self):
        metrics = []
        if self._reward_tracker is not None:
            metrics.append(self._reward_tracker)
        if self._compile_metrics is not None:
            metrics.append(self._compile_metrics)
        if self._compile_reward is not None:
            metrics.extend(self._compile_reward.metrics)
        metrics.extend(self._metrics)
        return metrics

    def _clear_previous_trainer_metrics(self):
        for module in self._flatten_modules(include_self=False):
            if not isinstance(module, Trainer):
                continue
            # A submodule might be a Trainer. In that case, we need to clear
            # the Trainer-related metrics, as they are not usable when a
            # new Trainer is instantiated.
            for m in self._get_own_metrics():
                module._tracker.untrack(m)
            module._reward_tracker = None
            module._compile_metrics = None
            if module._compile_reward is not None:
                module._compile_reward._metrics.clear()
            module._metrics.clear()

    @property
    def run_eagerly(self):
        return self._run_eagerly

    @run_eagerly.setter
    def run_eagerly(self, value):
        self._run_eagerly = value

    async def compute_reward(
        self,
        x=None,
        y=None,
        y_pred=None,
        training=True,
    ):
        """Compute the total reward, validate it, and return it.

        Subclasses can optionally override this method to provide custom reward
        computation logic.

        Args:
            x (list): Input data.
            y (list): Target data.
            y_pred (list): Predictions returned by the program (output of `program(x)`).
            training (bool): Whether we are training or evaluating the program.

        Returns:
            (float | None): The total reward as a scalar, or `None` if no reward results
                (which is the case when called by `Program.test_step`).
        """
        # The default implementation does not use `x` or `training`.
        del x
        del training
        rewards = []
        if self._compile_reward is not None:
            for y_t, y_p in zip(y, y_pred):
                reward = await self._compile_reward(y_t, y_p)
                if reward is not None:
                    rewards.append(reward)
        for reward in self.rewards:
            rewards.append(numpy.sum(reward))
        if len(rewards) == 1:
            total_reward = rewards[0]
        elif len(rewards) == 0:
            total_reward = numpy.zeros(())
        else:
            total_reward = numpy.mean(rewards)
        return float(total_reward)

    def stateless_compute_reward(
        self,
        trainable_variables,
        non_trainable_variables,
        metrics_variables,
        x=None,
        y=None,
        y_pred=None,
        training=True,
    ):
        var_mapping = list(zip(self.trainable_variables, trainable_variables))
        var_mapping.extend(zip(self.non_trainable_variables, non_trainable_variables))
        var_mapping.extend(zip(self.metrics_variables, metrics_variables))
        with backend.StatelessScope(state_mapping=var_mapping) as scope:
            # Note that this is needed for the regularization reward, which need
            # the latest value of train/non-trainable variables.
            reward = self._compute_reward(
                x,
                y,
                y_pred,
                training=training,
            )

        # Update non trainable vars (may have been updated in compute_reward)
        non_trainable_variables = []
        for v in self.non_trainable_variables:
            new_v = scope.get_current_value(v)
            non_trainable_variables.append(new_v)

        # Update metrics vars (may have been updated in compute_reward)
        metrics_variables = []
        for v in self.metrics_variables:
            new_v = scope.get_current_value(v)
            metrics_variables.append(new_v)
        return reward, (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        )

    async def compute_metrics(self, x, y, y_pred):
        """Update metric states and collect all metrics to be returned.

        Subclasses can optionally override this method to provide custom metric
        updating and collection logic. Custom metrics are not passed in
        `compile()`, they can be created in `__init__` or `build`. They are
        automatically tracked and returned by `self.metrics`.
        ```

        Args:
            x: Input data.
            y: Target data.
            y_pred: Predictions returned by the program output of `program.call(x)`.

        Returns:
            A `dict` containing values that will be passed to
                `synalinks.callbacks.CallbackList.on_train_batch_end()`. Typically,
                the values of the metrics listed in `self.metrics` are returned.
                Example: `{'reward': 0.2, 'accuracy': 0.7}`.
        """
        del x  # The default implementation does not use `x`.
        if self._compile_metrics is not None:
            for y_t, y_p in zip(y, y_pred):
                await self._compile_metrics.update_state(y_t, y_p)
        return self.get_metrics_result()

    def get_metrics_result(self):
        """Returns the program's metrics values as a dict.

        If any of the metric result is a dict (containing multiple metrics),
        each of them gets added to the top level returned dict of this method.

        Returns:
            (dict): A `dict` containing values of the metrics listed in `self.metrics`.
                Example: `{'reward': 0.2, 'accuracy': 0.7}`.
        """
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return python_utils.pythonify_logs(return_metrics)

    async def fit(
        self,
        x=None,
        y=None,
        batch_size=1,
        epochs=1,
        verbose="auto",
        callbacks=None,
        validation_split=0.1,
        validation_data=None,
        shuffle=True,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_batch_size=32,
        validation_freq=1,
    ):
        """Trains the program for a fixed number of epochs (dataset iterations).

        Args:
            x (np.ndarray | generator): Input data. It can be:
                - A NumPy array (or array-like), or a list of `DataModel` arrays
                    (in case the model has multiple inputs).
                - A list of dict mapping input names to the corresponding `DataModel`s,
                    if the program has named inputs.
                - A Python generator function yielding `(inputs, targets)`.
            y (np.ndarray): Target data. Like the input data `x`, it can be either NumPy
                array(s) of `DataModel`(s). If `x` is a Python generator function,
                `y` should not be specified since targets will be obtained from
                `x`.
            batch_size (int): Integer or `None`.
                Number of samples per batch of computation.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your input data `x` is a
                Python generator function since they generate batches.
            epochs (int): Integer. Number of epochs to train the program.
                An epoch is an iteration over the entire `x` and `y`
                data provided (unless the `steps_per_epoch` flag is set to
                something other than None).
                Note that in conjunction with `initial_epoch`,
                `epochs` is to be understood as "final epoch".
                The program is not trained for a number of iterations
                given by `epochs`, but merely until the epoch
                of index `epochs` is reached.
            verbose (int): `"auto"`, 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = one line per epoch.
                "auto" becomes 1 for most cases.
                Note that the progress bar is not
                particularly useful when logged to a file,
                so `verbose=2` is recommended when not running interactively
                (e.g., in a production environment). Defaults to `"auto"`.
            callbacks (list): List of `synalinks.callbacks.Callback` instances.
                List of callbacks to apply during training.
                See `synalinks.callbacks`. Note
                `synalinks.callbacks.ProgbarLogger` and
                `synalinks.callbacks.History` callbacks are created
                automatically and need not be passed to `program.fit()`.
                `synalinks.callbacks.ProgbarLogger` is created
                or not based on the `verbose` argument in `program.fit()`.
            validation_split (float): Float between 0 and 1.
                Fraction of the training data to be used as validation data.
                The program will set apart this fraction of the training data,
                will not train on it, and will evaluate the reward and any program
                metrics on this data at the end of each epoch. The validation
                data is selected from the last samples in the `x` and `y` data
                provided, before shuffling.
                This argument is only supported when `x` and `y` are made of
                data_models.
                If both `validation_data` and `validation_split` are provided,
                `validation_data` will override `validation_split`.
            validation_data (tuple | iterator): Data on which to evaluate
                the reward and any program metrics at the end of each epoch.
                The program will not be trained on this data.
                `validation_data` will override `validation_split`.
                It can be:
                - A tuple `(x_val, y_val)` of `DataModel`s lists.
            shuffle (bool): Whether to shuffle the training data before each
                epoch. This argument is ignored when `x` is a Python generator function.
            initial_epoch (int): Integer.
                Epoch at which to start training
                (useful for resuming a previous training run).
            steps_per_epoch (int): Integer or `None`.
                Total number of steps (batches of samples) before declaring one
                epoch finished and starting the next epoch. When training with
                input data_models arrays, the default `None` means that the
                value used is the number of samples in your dataset divided by
                the batch size, or 1 if that cannot be determined.
                If `x` is a Python generator function, the
                epoch will run until the input dataset is exhausted. When
                passing an infinitely repeating dataset, you must specify the
                `steps_per_epoch` argument, otherwise the training will run
                indefinitely.
            validation_steps (int): Integer or `None`.
                Only relevant if `validation_data` is provided.
                Total number of steps (batches of samples) to draw before
                stopping when performing validation at the end of every epoch.
                If `validation_steps` is `None`, validation will run until the
                `validation_data` dataset is exhausted. In the case of an
                infinitely repeating dataset, it will run indefinitely. If
                `validation_steps` is specified and only part of the dataset
                is consumed, the evaluation will start from the beginning of the
                dataset at each epoch. This ensures that the same validation
                samples are used every time.
            validation_batch_size (int): Integer or `None`.
                Number of samples per validation batch.
                If unspecified, will default to `batch_size`.
                Do not specify the `validation_batch_size` if your data is a
                `synalinks.utils.PyDataset`, `tf.data.Dataset`,
                `torch.utils.data.DataLoader` or Python generator function
                since they generate batches.
            validation_freq (int): Only relevant if validation data is provided.
                Specifies how many training epochs to run
                before a new validation run is performed,
                e.g. `validation_freq=2` runs validation every 2 epochs.

        Returns:
            (History): A `History` object. Its `History.history` attribute is
                a record of training reward values and metrics values
                at successive epochs, as well as validation reward values
                and validation metrics values (if applicable).
        """
        self._assert_compile_called("fit")
        self._eval_epoch_iterator = None
        val_y, val_y = None, None

        if validation_split and validation_data is None:
            # Create the validation data using the training data. Only supported
            # for numpy arrays.
            (x, y), validation_data = array_slicing.train_validation_split(
                (x, y), validation_split=validation_split
            )

        if validation_data is not None:
            (val_x, val_y) = data_adapter_utils.unpack_x_y(validation_data)
        # Create an iterator that yields batches of input/target data.
        epoch_iterator = EpochIterator(
            x=x,
            y=y,
            batch_size=batch_size,
            steps_per_epoch=steps_per_epoch,
            shuffle=False,
            steps_per_execution=self.steps_per_execution,
        )

        if not all(module.built for module in self._flatten_modules()):
            # Build the model on one batch of data.
            for _, data in epoch_iterator:
                data_batch = data[0]
                self._auto_build(
                    iterator=epoch_iterator,
                    data_batch=data_batch,
                )
                break
        epoch_iterator.reset()

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=epochs,
                steps=steps_per_epoch,
                program=self,
            )

        self.stop_training = False
        callbacks.on_train_begin()
        training_logs = None
        logs = {}
        initial_epoch = self._initial_epoch or initial_epoch

        if self.trainable_variables and isinstance(
            self.optimizer, optimizers_module.Optimizer
        ):
            await self.optimizer.on_train_begin(
                self.trainable_variables,
            )

        for epoch in range(initial_epoch, epochs):
            self.reset_metrics()

            if self.trainable_variables and isinstance(
                self.optimizer, optimizers_module.Optimizer
            ):
                await self.optimizer.on_epoch_begin(
                    epoch,
                    self.trainable_variables,
                )

            callbacks.on_epoch_begin(epoch)
            with epoch_iterator.catch_stop_iteration():
                for step, iterator in epoch_iterator:
                    data = iterator[0]
                    x_batch, y_batch = data_adapter_utils.unpack_x_y(data)

                    if self.trainable_variables and isinstance(
                        self.optimizer, optimizers_module.Optimizer
                    ):
                        await self.optimizer.on_batch_begin(
                            step,
                            epoch,
                            self.trainable_variables,
                        )

                    callbacks.on_train_batch_begin(step)

                    logs = await self.train_on_batch(
                        step=step,
                        x=x_batch,
                        y=y_batch,
                        val_x=val_x,
                        val_y=val_y,
                        return_dict=True,
                    )

                    val_logs = await self.evaluate(
                        x=val_x,
                        y=val_y,
                        batch_size=validation_batch_size or batch_size,
                        steps=validation_steps,
                        callbacks=callbacks,
                        _use_cached_eval_dataset=False,
                    )

                    if self.trainable_variables and isinstance(
                        self.optimizer, optimizers_module.Optimizer
                    ):
                        await self.optimizer.on_batch_end(
                            step,
                            epoch,
                            self.trainable_variables,
                        )

                    callbacks.on_train_batch_end(step, logs)
                    if self.stop_training:
                        break

            # Override with model metrics instead of last step logs if needed.
            epoch_logs = dict(self._get_metrics_result_or_logs(logs))

            # Run validation.
            if validation_data is not None and self._should_eval(epoch, validation_freq):
                # Create EpochIterator for evaluation and cache it.
                if getattr(self, "_eval_epoch_iterator", None) is None:
                    self._eval_epoch_iterator = EpochIterator(
                        x=val_x,
                        y=val_y,
                        batch_size=validation_batch_size or batch_size,
                        steps_per_execution=self.steps_per_execution,
                        steps_per_epoch=validation_steps,
                        shuffle=False,
                    )

                val_logs = await self.evaluate(
                    x=val_x,
                    y=val_y,
                    batch_size=validation_batch_size or batch_size,
                    steps=validation_steps,
                    callbacks=callbacks,
                    _use_cached_eval_dataset=True,
                )
                val_logs = {"val_" + name: val for name, val in val_logs.items()}
                epoch_logs.update(val_logs)

            if self.trainable_variables and isinstance(
                self.optimizer, optimizers_module.Optimizer
            ):
                await self.optimizer.on_epoch_end(
                    epoch,
                    self.trainable_variables,
                )

            callbacks.on_epoch_end(epoch, epoch_logs)
            training_logs = epoch_logs
            if self.stop_training:
                break

        # If _eval_epoch_iterator exists, delete it after all epochs are done.
        if getattr(self, "_eval_epoch_iterator", None) is not None:
            del self._eval_epoch_iterator

        if self.trainable_variables and isinstance(
            self.optimizer, optimizers_module.Optimizer
        ):
            await self.optimizer.on_train_end(self.trainable_variables)

        callbacks.on_train_end(logs=training_logs)
        return self.history

    async def evaluate(
        self,
        x=None,
        y=None,
        batch_size=32,
        verbose="auto",
        steps=None,
        callbacks=None,
        return_dict=True,
        **kwargs,
    ):
        """Returns the reward value & metrics values for the program in test mode.

        Computation is done in batches (see the `batch_size` arg.)

        Args:
            x (np.ndarray | generator): Input data. It can be:
                - A NumPy array (or array-like), or a list of `DataModel` arrays
                    (in case the model has multiple inputs).
                - A list of dict mapping input names to the corresponding `DataModel`s,
                    if the program has named inputs.
                - A Python generator function yielding `(inputs, targets)`.
            y (np.ndarray): Target data. Like the input data `x`, it can be either NumPy
                array(s) of `DataModel`(s). If `x` is a Python generator function,
                `y` should not be specified since targets will be obtained from
                `x`.
            batch_size (int): Integer or `None`.
                Number of samples per batch of computation.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your input data `x` is a
                Python generator function since they generate batches.
            verbose (int | str): `"auto"`, 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = single line.
                `"auto"` becomes 1 for most cases.
                Note that the progress bar is not
                particularly useful when logged to a file, so `verbose=2` is
                recommended when not running interactively
                (e.g. in a production environment). Defaults to `"auto"`.
            steps (int): Integer or `None`.
                Total number of steps (batches of samples) to draw before
                declaring the evaluation round finished. If `steps` is `None`,
                it will run until `x` is exhausted. In the case of an infinitely
                repeating dataset, it will run indefinitely.
            callbacks (list): List of `synalinks.callbacks.Callback` instances.
                List of callbacks to apply during evaluation.
            return_dict (bool): If `True`, reward and metric results are returned as a
                dict, with each key being the name of the metric.
                If `False`, they are returned as a list.

        Returns:
            (float | list | dict): Scalar test reward
                (if the program has a single output and no metrics)
                or list of scalars (if the program has multiple outputs
                and/or metrics). The attribute `program.metrics_names` will give you
                the display labels for the scalar outputs.
        """
        self._assert_compile_called("evaluate")
        use_cached_eval_dataset = kwargs.pop("_use_cached_eval_dataset", False)
        if kwargs:
            raise ValueError(f"Arguments not recognized: {kwargs}")
        # Create an iterator that yields batches of input/target data.
        if use_cached_eval_dataset:
            epoch_iterator = self._eval_epoch_iterator
        else:
            epoch_iterator = EpochIterator(
                x=x,
                y=y,
                batch_size=batch_size,
                steps_per_epoch=steps,
                shuffle=False,
                steps_per_execution=self.steps_per_execution,
            )

        if not all(module.built for module in self._flatten_modules()):
            # Build the model on one batch of data.
            for _, data in epoch_iterator:
                data_batch = data[0]
                self._auto_build(
                    iterator=epoch_iterator,
                    data_batch=data_batch,
                )
                break
        epoch_iterator.reset()

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=False,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=1,
                steps=epoch_iterator.num_batches,
                program=self,
            )

        self.stop_evaluating = False
        callbacks.on_test_begin()
        logs = {}
        self.reset_metrics()
        for step, iterator in epoch_iterator:
            callbacks.on_test_batch_begin(step)
            data = iterator[0]
            x_batch, y_batch = data_adapter_utils.unpack_x_y(data)
            logs = await self.test_on_batch(
                x=x_batch,
                y=y_batch,
                return_dict=True,
            )
            callbacks.on_test_batch_end(step, logs)
            if self.stop_evaluating:
                break
        logs = self.get_metrics_result()
        callbacks.on_test_end(logs)

        if return_dict:
            return logs
        return self._flatten_metrics_in_order(logs)

    async def predict(
        self, x, batch_size=None, verbose="auto", steps=None, callbacks=None
    ):
        """Generates output predictions for the input samples.

        Computation is done in batches. This method is designed for batch
        processing of large numbers of inputs. It is not intended for use inside
        of loops that iterate over your data and process small numbers of inputs
        at a time.

        For small numbers of inputs that fit in one batch,
        directly use `__call__()` for faster execution, e.g.,
        `program(x)`, or `program(x, training=False)` if you have modules
        that behave differently during inference.

        Args:
            x (np.ndarray | generator): Input data. It can be:
                - A NumPy array (or array-like), or a list of `DataModel` arrays
                    (in case the model has multiple inputs).
                - A list of dict mapping input names to the corresponding `DataModel`s,
                    if the program has named inputs.
                - A Python generator function yielding `(inputs, targets)`.
            batch_size (int): Integer or `None`.
                Number of samples per batch of computation.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your input data `x` is a
                `synalinks.utils.PyDataset`, `tf.data.Dataset`,
                `torch.utils.data.DataLoader` or Python generator function
                since they generate batches.
            verbose (int): `"auto"`, 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = single line.
                `"auto"` becomes 1 for most cases. Note that the progress bar
                is not particularly useful when logged to a file,
                so `verbose=2` is recommended when not running interactively
                (e.g. in a production environment). Defaults to `"auto"`.
            steps (int): Total number of steps (batches of samples) to draw before
                declaring the prediction round finished. If `steps` is `None`,
                it will run until `x` is exhausted. In the case of an infinitely
                repeating dataset, it will run indefinitely.
            callbacks (list): List of `synalinks.callbacks.Callback` instances.
                List of callbacks to apply during prediction.

        Returns:
            (list): `JsonDataModel` array(s) of predictions.
                If the pipeline failed, a None is added to the predictions.
        """
        # Create an iterator that yields batches of input data.
        epoch_iterator = EpochIterator(
            x=x,
            batch_size=batch_size,
            steps_per_epoch=steps,
            shuffle=False,
            steps_per_execution=self.steps_per_execution,
        )

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=1,
                steps=epoch_iterator.num_batches,
                model=self,
            )

        self.stop_predicting = False
        callbacks.on_test_begin()
        outputs = []
        for step, iterator in epoch_iterator:
            callbacks.on_predict_batch_begin(step)
            data = iterator[0]
            x_batch, _ = data_adapter_utils.unpack_x_y(data)
            batch_outputs = await self.predict_on_batch(x_batch)
            outputs.extend(batch_outputs)
            callbacks.on_predict_batch_end(step, {"outputs": batch_outputs})
            if self.stop_predicting:
                break
        callbacks.on_predict_end()
        return np.array(outputs, dtype="object")

    async def train_on_batch(
        self,
        step,
        x,
        y=None,
        val_x=None,
        val_y=None,
        return_dict=False,
    ):
        """Runs a single optimization step on a single batch of data.

        Args:
            step (int): The training step.
            x (np.ndarray): Input data. Must be array-like.
            y (np.ndarray): Target data. Must be array-like.
            val_x (np.ndarray): Input validation data. Must be array-like.
            val_y (np.ndarray): Target validation data. Must be array-like.
            return_dict (bool): If `True`, reward and metric results are returned as a
                dict, with each key being the name of the metric. If `False`,
                they are returned as a list.

        Returns:
            (float | list | dict): A scalar reward value
                (when no metrics and `return_dict=False`), a list of reward
                and metric values (if there are metrics and `return_dict=False`),
                or a dict of metric and reward values (if `return_dict=True`).
        """
        if self.trainable_variables and isinstance(
            self.optimizer, optimizers_module.Optimizer
        ):
            metrics = await self.optimizer.optimize(
                step,
                self.trainable_variables,
                x=x,
                y=y,
                val_x=val_x,
                val_y=val_y,
            )
        else:
            warnings.warn("The program does not have any trainable variables.")
            y_pred = await self.predict_on_batch(val_x)
            reward = await self.compute_reward(
                x=val_x,
                y=val_y,
                y_pred=y_pred,
            )
            await self._reward_tracker.update_state(reward)
            metrics = await self.compute_metrics(val_x, val_y, y_pred)

        if return_dict:
            return metrics
        return self._flatten_metrics_in_order(metrics)

    async def test_on_batch(
        self,
        x,
        y=None,
        return_dict=False,
    ):
        """Test the program on a single batch of samples.

        Args:
            x (np.ndarray): Input data. Must be array-like.
            y (np.ndarray): Target data. Must be array-like.
            return_dict (bool): If `True`, reward and metric results are returned as a
                dict, with each key being the name of the metric. If `False`,
                they are returned as a list.

        Returns:
            (float | list | dict): A scalar reward value
                (when no metrics and `return_dict=False`), a list of reward
                and metric values (if there are metrics and `return_dict=False`),
                or a dict of metric and reward values (if `return_dict=True`).
        """
        y_pred = await self.predict_on_batch(x)

        reward = await self.compute_reward(
            x=x,
            y=y,
            y_pred=y_pred,
            training=False,
        )
        await self._reward_tracker.update_state(reward)

        metrics = await self.compute_metrics(x, y, y_pred)

        if return_dict:
            return metrics
        return self._flatten_metrics_in_order(metrics)

    async def predict_on_batch(self, x, training=False):
        """Returns predictions for a single batch of samples.

        Args:
            x (np.ndarray): Input data. Must be array-like.
            training (bool): Boolean. True if training.

        Returns:
            (list): list(s) of JsonDataModel predictions.
        """
        tasks = []
        for inputs in x:
            tasks.append(self(inputs, training=training))
        y_pred = await asyncio.gather(*tasks)
        return y_pred

    def get_compile_config(self):
        """Returns a serialized config with information for compiling the program.

        This method returns a config dictionary containing all the information
        (optimizer, reward, metrics, etc.) with which the program was compiled.

        Returns:
            (dict): A dict containing information for compiling the program.
        """
        if self.compiled and hasattr(self, "_compile_config"):
            return self._compile_config.serialize()

    def compile_from_config(self, config):
        """Compiles the program with the information given in config.

        This method uses the information in the config (optimizer, reward,
        metrics, etc.) to compile the program.

        Args:
            config (dict): Dict containing information for compiling the program.
        """
        has_overridden_compile = self.__class__.compile != Trainer.compile
        if has_overridden_compile:
            warnings.warn(
                "`compile()` was not called as part of program loading "
                "because the program's `compile()` method is custom. "
                "All subclassed Models that have `compile()` "
                "overridden should also override "
                "`get_compile_config()` and `compile_from_config(config)`. "
                "Alternatively, you can "
                "call `compile()` manually after loading.",
                stacklevel=2,
            )
            return
        config = serialization_lib.deserialize_synalinks_object(config)
        self.compile(**config)
        if hasattr(self, "optimizer") and self.built:
            # Create optimizer variables/programs.
            if not self.optimizer.built:
                run_maybe_nested(self.optimizer.build(self.trainable_variables))

    def _should_reward(self, epoch, validation_freq):
        epoch = epoch + 1  # one-index the user-facing epoch.
        if isinstance(validation_freq, int):
            return epoch % validation_freq == 0
        elif isinstance(validation_freq, list):
            return epoch in validation_freq
        else:
            raise ValueError(
                "Expected `validation_freq` to be a list or int. "
                f"Received: validation_freq={validation_freq} of the "
                f"type {type(validation_freq)}."
            )

    def _get_metrics_result_or_logs(self, logs):
        """Returns program metrics as a dict if the keys match with input logs.

        When the training / evaluation is performed with an asynchronous steps,
        the last scheduled `train / test_step` may not give the latest metrics
        because it is not guaranteed to be executed the last. This method gets
        metrics from the program directly instead of relying on the return from
        last step function.

        When the user has custom train / test step functions, the metrics
        returned may be different from `Program.metrics`. In those instances,
        this function will be no-op and return the logs passed in.

        Args:
            logs (dict): A `dict` of metrics returned by train / test step function.

        Returns:
            (dict): A `dict` containing values of the metrics listed in `self.metrics`
                when logs and program metrics keys match. Otherwise it returns input
                `logs`.
        """
        metric_logs = self.get_metrics_result()
        # Verify that train / test step logs passed and metric logs have
        # matching keys. It could be different when using custom step functions,
        # in which case we return the logs from the last step.
        if isinstance(logs, dict) and set(logs.keys()) == set(metric_logs.keys()):
            return metric_logs
        return logs

    def _flatten_metrics_in_order(self, logs):
        """Turns `logs` dict into a list as per key order of `metrics_names`."""
        metric_names = []
        for metric in self.metrics:
            if isinstance(metric, CompileMetrics):
                metric_names += [sub_metric.name for sub_metric in metric.metrics]
            else:
                metric_names.append(metric.name)
        results = []
        for name in metric_names:
            if name in logs:
                results.append(logs[name])
        for key in sorted(logs.keys()):
            if key not in metric_names:
                results.append(logs[key])
        if len(results) == 1:
            return results[0]
        return results

    def _assert_compile_called(self, method_name=None):
        if not self.compiled:
            msg = "You must call `compile()` before "
            if metrics_module:
                msg += "using the program."
            else:
                msg += f"calling `{method_name}()`."
            raise ValueError(msg)

    def _auto_build(self, iterator=None, data_batch=None):
        program_unbuilt = not all(module.built for module in self._flatten_modules())
        compile_metrics_unbuilt = (
            self._compile_metrics is not None and not self._compile_metrics.built
        )
        compile_reward_unbuilt = (
            self._compile_reward is not None and not self._compile_reward.built
        )
        optimizer_unbuilt = self.optimizer is not None and not self.optimizer.built
        if program_unbuilt or compile_metrics_unbuilt or compile_reward_unbuilt:
            if data_batch is None:
                for _, data_or_iterator in iterator:
                    if isinstance(data_or_iterator, (list, tuple)):
                        data_batch = data_or_iterator[0]
                    else:
                        data_batch = next(data_or_iterator)
                    break
            (x, y) = data_batch
            try:
                y_pred = run_maybe_nested(self.predict_on_batch(x))
            except Exception as e:
                raise RuntimeError(
                    "Unable to automatically build the program. "
                    "Please build it yourself before calling "
                    "fit/evaluate/predict. "
                    "A program is 'built' when its variables have "
                    "been created and its `self.built` attribute "
                    "is True. Usually, calling the program on a batch "
                    "of data is the right way to build it.\n"
                    "Exception encountered:\n"
                    f"'{e}'"
                )
            if compile_metrics_unbuilt:
                # Build all metric state with `backend.compute_output_spec`.
                run_maybe_nested(
                    backend.compute_output_spec(
                        self.compute_metrics,
                        x,
                        y,
                        y_pred,
                    )
                )
            if compile_reward_unbuilt:
                # Build `CompileReward` state with `backend.compute_output_spec`.
                run_maybe_nested(
                    backend.compute_output_spec(
                        self.compute_reward,
                        x,
                        y,
                        y_pred,
                        training=False,
                    )
                )
        if optimizer_unbuilt:
            # Build optimizer
            run_maybe_nested(self.optimizer.build(self.trainable_variables))
        self._post_build()

    def _assert_compile_called(self, method_name=None):
        if not self.compiled:
            msg = "You must call `compile()` before "
            if metrics_module:
                msg += "using the model."
            else:
                msg += f"calling `{method_name}()`."
            raise ValueError(msg)

    def _should_eval(self, epoch, validation_freq):
        epoch = epoch + 1  # one-index the user-facing epoch.
        if isinstance(validation_freq, int):
            return epoch % validation_freq == 0
        elif isinstance(validation_freq, list):
            return epoch in validation_freq
        else:
            raise ValueError(
                "Expected `validation_freq` to be a list or int. "
                f"Received: validation_freq={validation_freq} of the "
                f"type {type(validation_freq)}."
            )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import os


def get_fake_data(filename):
    from synalinks.cli import datasets

    folder_path = datasets.__file__[: -len("/__init__.py")]
    json_data = ""
    with open(os.path.join(folder_path, filename), "r") as f:
        json_data = f.read()
    return json_data


PROJECT_CONFIG_FILENAME = "synalinks_project.json"

TEMPLATE_CONFIG_FILENAME = "synalinks_template.json"

README_TEMPLATE = """# {{config.project_name}}
## {{config.project_description}}

![{{config.project_name}}]({{config.project_name}}.png)

![training_history](training_history.png)

# Install dependencies

To plot your programs, you will need graphviz

```shell
sudo apt update
sudo apt install graphviz
```

Create your virtual environment and install synalinks

```shell
cd {{config.project_name}}
uv venv
uv pip install synalinks
```

# Train your application

```shell
cd {{config.project_name}}
python3 scripts/train.py
```

# Serve your API

```shell
cd {{config.project_name}}
docker compose up --build
```

Note: Ollama models are great for experimenting/training but not to use in production.

Powered by [🧠🔗 - synalinks](https://github.com/SynaLinks/synalinks)

More information available in the [documentation](https://synalinks.github.io/synalinks/)

If you have questions, check out the [FAQ](https://synalinks.github.io/synalinks/FAQ/).

Join our [Discord](https://discord.gg/82nt97uXcM) to never miss any update!

Happy coding 😀!
"""

PYPROJECT_TEMPLATE = """[project]
name = "{{config.package_name}}"
description = "{{config.project_description}}"
readme = "README.md"
requires-python = ">=3.9"
dynamic = ["version"]
classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: 3 :: Only",
    "Operating System :: Unix",
    "Operating System :: MacOS",
]
dependencies = [
    "synalinks"
]
[tool.setuptools.dynamic]
version = {attr = "{{config.package_name}}.src.version.__version__"}

[tool.setuptools.packages.find]
include = ["{{config.package_name}}.src", "{{config.package_name}}.src.*"]
"""


GITIGNORE_TEMPLATE = """# Python-generated files
__pycache__/
*.py[oc]
build/
dist/
wheels/
*.egg-info

# Virtual environments
.venv

# Environment variables
.env
"""

MAIN_TEMPLATE = """import os
import synalinks
import psutil
import asyncio
from dotenv import load_dotenv
from fastapi import FastAPI
# Uncomment for streaming apps
# from fastapi.responses import StreamingResponse

# Import your input data model for your API
from {{config.package_name}}.src.data_models import Query
# Import your custom module
from {{config.package_name}}.src.modules import AnswerWithChainOfThought

# Load the .env variables
load_dotenv()

# Clear Synalinks context
synalinks.clear_session()

async def create_program():
    language_model = synalinks.LanguageModel(
        model=os.environ.get(
            "LANGUAGE_MODEL",
            "{{config.model_provider}}/{{config.language_model}}",
        ),
        api_base=os.environ.get("MODEL_API_BASE", None),
    )
        
    inputs = synalinks.Input(data_model=Query)
    outputs = await AnswerWithChainOfThought(
        language_model=language_model,
    )(inputs)

    return synalinks.Program(
        inputs=inputs,
        outputs=outputs,
        name="{{config.package_name}}",
        description="{{config.project_description}}",
    )
    
program = asyncio.run(create_program())

# Load your application variables (the ones optimized by the training)
program_variables_filepath = "checkpoint.program.variables.json"
if os.path.exists(program_variables_filepath):
    program.load_variables(program_variables_filepath)
    
# Setup FastAPI
app = FastAPI()

@app.get("/health")
def health_check():
    return {"status": "healthy"}


@app.get("/system_check")
def system_check():
    cpu_usage = psutil.cpu_percent(interval=1)
    memory_info = psutil.virtual_memory()
    disk_usage = psutil.disk_usage("/")
    return {
        "cpu_usage": cpu_usage,
        "memory_usage": memory_info.percent,
        "disk_usage": disk_usage.percent,
    }


@app.post("/v1/{{config.package_name}}")
async def {{config.package_name}}(inputs: Query):
    result = await program(inputs)
    # Uncomment for streaming apps
    # return StreamingResponse(result, media_type="application/json") if result else None
    return result.get_json() if result else None
"""

DOCKERFILE_TEMPLATE = """FROM python:3.13

COPY ./requirements.txt ./requirements.txt

RUN apt update && apt install -y graphviz

RUN pip install --no-cache-dir --upgrade -r ./requirements.txt

COPY . ./

CMD ["fastapi", "run", "{{config.package_name}}/src/main.py", "--port", "8000"]
"""

DOCKER_COMPOSE_TEMPLATE = """services:
  {{config.package_name}}:
    build:
      context: .
      dockerfile: Dockerfile
    ports:
      - "8000:8000"
    env_file:
      - .env
    extra_hosts:
      - "host.docker.internal:host-gateway"
"""

VERSION_TEMPLATE = """
# Unique source of truth for the version number.
__version__ = "0.1.0"
"""

INIT_TEMPLATE = """from {{config.package_name}}.src.version import __version__
from {{config.package_name}}.src.main import main
"""

DATA_MODELS_INIT_TEMPLATE = """from {{config.package_name}}.src.data_models.query import Query
from {{config.package_name}}.src.data_models.answer import Answer
"""

MODULES_INIT_TEMPLATE = """from {{config.package_name}}.src.modules.answer_with_chain_of_thought import AnswerWithChainOfThought
"""

REQUIREMENTS_TEMPLATE = """fastapi[standard]
psutil
synalinks
"""

QUERY_TEMPLATE = """import synalinks

class Query(synalinks.DataModel):
    query: str = synalinks.Field(
        description="The user query",
    )
"""

ANSWER_TEMPLATE = """import synalinks

class Answer(synalinks.DataModel):
    answer: str = synalinks.Field(
        description="The correct answer",
    )
"""

PROGRAM_TEMPLATE = """import synalinks
from {{config.package_name}}.src.data_models import Answer

class AnswerWithChainOfThought(synalinks.Program):
    \"""Answer step by step.
    
    Args:
        language_model (LanguageModel): The language model to use.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    \"""

    def __init__(
        self,
        language_model=None,
        name=None,
        description=None,
        trainable=None,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        self.language_model = language_model

    async def build(self, inputs):
        outputs = await synalinks.ChainOfThought(
            data_model=Answer,
            language_model=self.language_model,
        )(inputs)
        
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            name=self.name,
            description=self.description,
            trainable=self.trainable,
        )
"""

TRAIN_TEMPLATE = """#!/usr/bin/python3
import os
import synalinks
import asyncio
import json
import numpy as np
from {{config.package_name}}.src.data_models import Query
from {{config.package_name}}.src.data_models import Answer
from {{config.package_name}}.src.modules import AnswerWithChainOfThought

# Clear Synalinks context
synalinks.clear_session()

# The training is done in batch, the batch size specify the 
# number of parralel program execution performed to train 
# the application. The reward is averaged per batch yielding 
# a better estimation of your program success.
BATCH_SIZE=32

# The epochs refer to the number of time the whole dataset is
# proccessed. At the end of each epochs, the optimization is
# performed. So the epochs is the number of successive optimization.
EPOCHS=4

def load_dataset(
    input_data_model,
    output_data_model,
    x_train_filepath="datasets/x_train.json",
    y_train_filepath="datasets/y_train.json",
    x_test_filepath="datasets/x_test.json",
    y_test_filepath="datasets/y_test.json",
):
    x_train = []
    y_train = []
    x_test = []
    y_test = []
    
    with open(x_train_filepath, "r") as f:
        json_data = json.loads(f.read())
        for data_point in json_data:
            x_train.append(input_data_model(**data_point))
    
    with open(y_train_filepath, "r") as f:
        json_data = json.loads(f.read())
        for data_point in json_data:
            y_train.append(output_data_model(**data_point))
        
    with open(x_test_filepath, "r") as f:
        json_data = json.loads(f.read())
        for data_point in json_data:
            x_test.append(input_data_model(**data_point))
            
    with open(y_test_filepath, "r") as f:
        json_data = json.loads(f.read())
        for data_point in json_data:
            y_test.append(output_data_model(**data_point))
    
    # Convert the dataset into numpy arrays
    
    x_train = np.array(x_train, dtype="object")
    y_train = np.array(y_train, dtype="object")
    
    x_test = np.array(x_test, dtype="object")
    y_test = np.array(y_test, dtype="object")

    return (x_train, y_train), (x_test, y_test)

async def train_program():
    language_model = synalinks.LanguageModel(
        model=os.environ.get(
            "LANGUAGE_MODEL",
            "{{config.model_provider}}/{{config.language_model}}"
        ),
    )
    
    embedding_model = synalinks.EmbeddingModel(
        model=os.environ.get(
            "EMBEDDING_MODEL",
            "{{config.model_provider}}/{{config.embedding_model}}"
        ),
    )
    
    inputs = synalinks.Input(data_model=Query)
    outputs = await AnswerWithChainOfThought(
        language_model=language_model,
    )(inputs)

    program = synalinks.Program(
        inputs=inputs,
        outputs=outputs,
        name="{{config.project_name}}",
        description="{{config.project_description}}",
    )
    
    embedding_model = synalinks.EmbeddingModel(
        model=os.environ.get(
            "EMBEDDING_MODEL",
            "{{config.model_provider}}/{{config.embedding_model}}",
        ),
    )
    
    synalinks.utils.plot_program(
        program,
        show_module_names=True,
        show_schemas=True,
    )
    
    checkpoint_filepath = "checkpoint.program.variables.json"
    
    program_checkpoint_callback = synalinks.callbacks.ProgramCheckpoint(
        filepath=checkpoint_filepath,
        monitor="val_reward",
        save_variables_only=True,
        mode="max",
        save_best_only=True,
    )
    
    program.compile(
        reward=synalinks.rewards.CosineSimilarity(
            # Filter to keep only the `answer` field in order to compute the reward
            in_mask=["answer"],
            # The embedding model to use to compute the similarity
            embedding_model=embedding_model,
        ),
        optimizer=synalinks.optimizers.RandomFewShot(),
        metrics=[
            synalinks.metrics.F1Score(in_mask=["answer"]),
        ],
    )
    
    (x_train, y_train), (x_test, y_test) = load_dataset(Query, Answer)
    
    history = await program.fit(
        x=x_train,
        y=y_train,
        validation_data=(x_test, y_test),
        batch_size=BATCH_SIZE,
        epochs=EPOCHS,
        callbacks=[program_checkpoint_callback],
    )
    
    synalinks.utils.plot_history(
        history,
    )
    
def main():
    asyncio.run(train_program())

if __name__ == "__main__":
    main()
"""

PROJECT_STRUCTURE_TEMPLATE = {
    "README.md": README_TEMPLATE,
    "pyproject.toml": PYPROJECT_TEMPLATE,
    "{{config.package_name}}": {
        "src": {
            "data_models": {
                "__init__.py": DATA_MODELS_INIT_TEMPLATE,
                "query.py": QUERY_TEMPLATE,
                "answer.py": ANSWER_TEMPLATE,
                "README.md": (
                    "For the list of built-in data models check the documentation "
                    "[here](https://synalinks.github.io/synalinks/Synalinks%20API/Data%20Models%20API/The%20Base%20DataModels/)",
                ),
            },
            "modules": {
                "__init__.py": MODULES_INIT_TEMPLATE,
                "answer_with_chain_of_thought.py": PROGRAM_TEMPLATE,
                "README.md": (
                    "For the list of built-in modules check the documentation "
                    "[here](https://synalinks.github.io/synalinks/Synalinks%20API/Modules%20API/)",
                ),
            },
            "tools": {
                "__init__.py": "",
                "README.md": "Here lies the tools for your agents",
            },
            "__init__.py": "",
            "version.py": VERSION_TEMPLATE,
            "main.py": MAIN_TEMPLATE,
        },
    },
    "scripts": {
        "train.py": TRAIN_TEMPLATE,
        "README.md": "Here lies your scripts",
    },
    "datasets": {
        "x_train.json": get_fake_data("x_train.json"),
        "y_train.json": get_fake_data("y_train.json"),
        "x_test.json": get_fake_data("x_test.json"),
        "y_test.json": get_fake_data("y_test.json"),
        "README.md": "Here lies your custom datasets",
    },
    "requirements.txt": REQUIREMENTS_TEMPLATE,
    ".gitignore": GITIGNORE_TEMPLATE,
    "Dockerfile": DOCKERFILE_TEMPLATE,
    "docker-compose.yml": DOCKER_COMPOSE_TEMPLATE,
}



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.hooks.hook import Hook as Hook
from synalinks.src.hooks.hook_list import HookList as HookList
from synalinks.src.hooks.logger import Logger as Logger
from synalinks.src.hooks.monitor import Monitor as Monitor
